from enum import Enum
from tabulate import tabulate

from ahc.Ahc import ComponentModel, Event, EventTypes, GenericMessageHeader, GenericMessage, MessageDestinationIdentifiers
from threading import Timer

class DsdvTimer():

    def __init__(self):
        self.working = False

    def start(self, interval, handleFunction, args=None, kwargs=None):
        self.interval = interval
        self.handleFunction = handleFunction
        self.thread = Timer(self.interval, self.handleFunction, args, kwargs)
        self.thread.daemon = True
        self.thread.start()
        self.working = True

    def cancel(self):
        self.thread.cancel()
        self.working = False


# Definition of message types in the DSDV routing protocol
# Full Dump Update Message: Entire routing table will be send
# Incremental Update: One entry in the routing table will be send
class DsdvMessageTypes(Enum): # Control Plane message types
    fullDumpUpdate = "Full Dump Update Message"
    incrementalUpdate = "Incremental Update Message"


class DataMessageTypes(Enum):
    appData = "Application Data Message"


# Definition of the index for each column in the routing table
class RoutingTableColumn(Enum):
    destination = 0
    nextHop = 1
    metric = 2
    sequenceNumber = 3


class DsdvService(ComponentModel):

    def __init__(self, componentname, componentinstancenumber):
        super().__init__(componentname, componentinstancenumber)    # init super class
        self.routingTable = []      # a routing table to keep minimum distance information to the other nodes
        self.sequenceNumber = 0     # sequence number of the broadcast messages for self component

        # First entry to add to self routing table is the self distance
        self.routingTable.append([componentinstancenumber, componentinstancenumber, 0, self.sequenceNumber])
        # Create a timer thread instance to send periodic update messages
        self.periodicUpdateTimer = DsdvTimer()
        # First message will be send 1 second after the init. The period will be different
        self.periodicUpdateTimer.start(1, self.send_periodic_update)
        # Create timer thread instances to be able to schedule update messages aside from the periodic update
        self.scheduleFullDumpUpdate = DsdvTimer()
        self.scheduleIncrementalUpdate = DsdvTimer()
        # Decide interval of the periodic update messages. No hard limitation
        self.periodicUpdateMessageInterval = 10

        # Added only to debug. The variables can be removed
        self.schedulePrintRoutingTable = DsdvTimer()
        self.numberOfIncrementalUpdate = 0
        self.numberOfFullDumpUpdate = 0
        self.numberOfUpdateMessage = 0
        self.printRoutingTableOnPeriodicUpdates = True
        self.printNumberOfMessagesOnPeriodicUpdates = False
        self.printNumberOfMessagesOnEveryUpdate = False

    # This function increases the self sequence number and
    # updates the routing table according to new self sequence number
    #
    def increase_self_sequence_number(self):
        # Sequence numbers should be even numbers
        self.sequenceNumber = self.sequenceNumber + 2
        # Update the routing table according to new self sequence number
        for row in self.routingTable:
            if row[RoutingTableColumn.destination.value] == self.componentinstancenumber:
                row[RoutingTableColumn.sequenceNumber.value] = self.sequenceNumber

    # This function generates a full dump update message and sends to lower layer
    # @param routingTable is the self routing table to broadcast
    #
    def send_full_dump_update(self, routingTable):
        payload = routingTable.copy()
        destination = MessageDestinationIdentifiers.LINKLAYERBROADCAST
        nexthop = MessageDestinationIdentifiers.LINKLAYERBROADCAST
        header = GenericMessageHeader(DsdvMessageTypes.fullDumpUpdate, self.componentinstancenumber, destination,
                                      nexthop)

        message = GenericMessage(header, payload)
        self.send_down(Event(self, EventTypes.MFRT, message))
        self.increase_self_sequence_number()

        # Added to debug
        self.numberOfFullDumpUpdate = self.numberOfFullDumpUpdate + 1
        self.numberOfUpdateMessage = self.numberOfUpdateMessage + 1
        if self.printNumberOfMessagesOnEveryUpdate:
            print(self.unique_name() + " Number Of Update Message: " + str(self.numberOfUpdateMessage))

    # This function generates an incremental update message and sends to lower layer
    # @param routingTableEntry is an entry of the self routing table to broadcast
    #
    def send_incremental_update(self, routingTableEntry):
        payload = routingTableEntry.copy()
        destination = MessageDestinationIdentifiers.LINKLAYERBROADCAST
        nexthop = MessageDestinationIdentifiers.LINKLAYERBROADCAST
        header = GenericMessageHeader(DsdvMessageTypes.incrementalUpdate, self.componentinstancenumber, destination,
                                      nexthop)

        message = GenericMessage(header, payload)
        self.send_down(Event(self, EventTypes.MFRT, message))
        self.increase_self_sequence_number()

        # Added to debug
        self.numberOfIncrementalUpdate = self.numberOfIncrementalUpdate + 1
        self.numberOfUpdateMessage = self.numberOfUpdateMessage + 1
        if self.printNumberOfMessagesOnEveryUpdate:
            print(self.unique_name() + " Number Of Update Message: " + str(self.numberOfUpdateMessage))

    # This function provides a periodic call of the send_full_dump_update() function with the help of a timer thread
    #
    def send_periodic_update(self):
        self.send_full_dump_update(self.routingTable)
        self.periodicUpdateTimer.cancel()
        self.periodicUpdateTimer.start(self.periodicUpdateMessageInterval, self.send_periodic_update)

        # added to debug purpose only
        if self.printRoutingTableOnPeriodicUpdates:
            self.schedulePrintRoutingTable.start(self.componentinstancenumber, self.print_self_routing_table)
        elif self.printNumberOfMessagesOnPeriodicUpdates:
            print(self.unique_name() + " Number Of Update Message: " + str(self.numberOfUpdateMessage))

    # This function sends full dump update message as a result of the scheduling
    #
    def send_scheduled_full_dump_update(self, routingTable):
        if self.scheduleFullDumpUpdate.working:
            self.scheduleFullDumpUpdate.cancel()
            self.send_full_dump_update(routingTable)

    # This function sends incremental update message as a result of the scheduling
    #
    def send_scheduled_incremental_update(self, routingTableEntry):
        if self.scheduleIncrementalUpdate.working:
            self.scheduleIncrementalUpdate.cancel()
            self.send_incremental_update(routingTableEntry)

    # This function schedules sending an incremental update message
    # The function invoked from processing an update message
    #
    def schedule_incremental_update(self, routingTableEntry):
        # If a full dump update message is scheduled, there is no need to schedule an incremental update since
        # the full dump update message covers the incremental update
        if self.scheduleFullDumpUpdate.working:
            pass
        else:
            # If more than 1 request exist for incremental update, schedule a full dump update instead
            if self.scheduleIncrementalUpdate.working:
                self.scheduleIncrementalUpdate.cancel()
                self.scheduleFullDumpUpdate.start(1, self.send_scheduled_full_dump_update, [self.routingTable])
            else:
                self.scheduleIncrementalUpdate.start(2, self.send_scheduled_incremental_update, [routingTableEntry])

    # This function schedules sending a full dump update message
    # The function invoked from processing an update message
    #
    def schedule_full_dump_update(self):
        if self.scheduleFullDumpUpdate.working == False:
            self.scheduleFullDumpUpdate.start(1, self.send_scheduled_full_dump_update, [self.routingTable])

    # This function gets incoming full dump update messages and process
    # Process means update self routing table according to the incoming messages
    # @param payload is the payload of the incoming message which shows the entire routing table of the one of
    # the neighbors
    # @param messageFrom shows source address of the incoming message
    #
    def process_full_dump_update_message(self, payload, messageFrom):
        updatedRows = []    # Keep record of the updated entries in the self routing table

        # Process every entry in the incoming message
        for rowInPayload in payload:
            singleEntry = rowInPayload.copy()
            # Increment the metric value by 1, since the distance information was according to my neighbor
            singleEntry[RoutingTableColumn.metric.value] = singleEntry[RoutingTableColumn.metric.value] + 1

            isRoutingTableUpdated = self.process_single_entry_in_message(singleEntry, messageFrom)
            if isRoutingTableUpdated == True:
                updatedRows.append(singleEntry)

        if len(updatedRows) > 1:   # More than 1 entry changed. Invoke to schedule full dump update message
            self.schedule_full_dump_update()
        elif len(updatedRows) == 1:    # Only 1 entry changed. Invoke to schedule incremental update message is enough
            self.schedule_incremental_update(updatedRows[0])

    # This function gets incoming incremental update messages and process
    # Process means update self routing table according to the incoming messages
    # @param payload is the payload of the incoming message which shows single entry of the routing table of the one
    # of the neighbors
    # @param messageFrom shows the source address of the incoming message
    #
    def process_incremental_update_message(self, payload, messageFrom):
        singleEntry = payload.copy()
        # Increment the metric value by 1, since the distance information was according to my neighbor
        singleEntry[RoutingTableColumn.metric.value] = singleEntry[RoutingTableColumn.metric.value] + 1
        isRoutingTableUpdated = self.process_single_entry_in_message(singleEntry, messageFrom)

        if isRoutingTableUpdated == True:
            self.schedule_incremental_update(singleEntry)

    # This function processes a single entry of the incoming update message
    # @param singleEntry is the entry to process
    # @param messageFrom shows the source address of the incoming message
    # @return if the self routing table is changed or not. Only the crucial updates considered as change
    #
    def process_single_entry_in_message(self, singleEntry, messageFrom):
        destColumnIndex = RoutingTableColumn.destination.value
        nextHopColumnIndex = RoutingTableColumn.nextHop.value
        seqNumberColumnIndex = RoutingTableColumn.sequenceNumber.value
        metricColumnIndex = RoutingTableColumn.metric.value
        crucialUpdateExist = False


        # -----Update the entry of the incoming message according to self component---
        # Next hop information will be sender of the message
        singleEntry[nextHopColumnIndex] = messageFrom
        # ------------------

        # If the entry is not about myself, continue the process
        if singleEntry[destColumnIndex] != self.componentinstancenumber:
            # Search the entry which has same destination node information in self routing table
            rowIndex = self.search_for_entry(singleEntry[destColumnIndex])

            if rowIndex == -1:  # No match in self.routing table
                self.routingTable.append(singleEntry)
                self.routingTable.sort()
                crucialUpdateExist = True

            else:
                # If the entry in the message has a bigger sequence number, update the information
                if singleEntry[seqNumberColumnIndex] > self.routingTable[rowIndex][seqNumberColumnIndex]:
                    self.routingTable[rowIndex] = singleEntry
                    crucialUpdateExist = self.decide_which_update_is_crucial(singleEntry, self.routingTable[rowIndex])
                # If the entry in the message has same sequence number, choose the lower metric value
                elif singleEntry[seqNumberColumnIndex] == self.routingTable[rowIndex][seqNumberColumnIndex]:
                    if singleEntry[metricColumnIndex] < self.routingTable[rowIndex][metricColumnIndex]:
                        self.routingTable[rowIndex] = singleEntry
                        crucialUpdateExist = self.decide_which_update_is_crucial(singleEntry,
                                                                                 self.routingTable[rowIndex])
        return crucialUpdateExist

    # This function searches for an entry in the self routing table
    # Function finds the entry which caries the information about the given destination node
    # @param destinationNode is the node to look for in the self routing table
    # @return rowIndex is the index of the row which carries the information about the given destination node
    #   - If the information is not found, return -1
    #
    def search_for_entry(self, destinationNode):
        destColumnIndex = RoutingTableColumn.destination.value
        rowIndex = 0

        for rowInSelfTable in self.routingTable:
            if rowInSelfTable[destColumnIndex] == destinationNode:
                return rowIndex
            else:
                rowIndex = rowIndex + 1

        return -1

    # If the change in the entry only the sequence number, in other words the metric and next hop
    # information is same, do not broadcast the information. No significant change
    #
    def decide_which_update_is_crucial(self, entryInMessage, entryInSelfTable):

        nextHopColumnIndex = RoutingTableColumn.nextHop.value
        metricColumnIndex = RoutingTableColumn.metric.value

        if entryInMessage[metricColumnIndex] != entryInSelfTable[metricColumnIndex] or \
                entryInMessage[nextHopColumnIndex] != entryInMessage[nextHopColumnIndex]:
            return True
        else:
            return False

    def print_self_routing_table(self):
        print("\r\n\r\n" + self.unique_name() + " Routing Table")
        print(tabulate(self.routingTable, headers=[RoutingTableColumn.destination.name, RoutingTableColumn.nextHop.name,
                                                   RoutingTableColumn.metric.name,
                                                   RoutingTableColumn.sequenceNumber.name], tablefmt='grid'))

    # This function handles the messages coming from lower layer
    # Function overrides super class function
    #
    def on_message_from_bottom(self, eventobj: Event):
        msg = eventobj.eventcontent
        hdr = msg.header
        payload = msg.payload

        #----Control Plane messages---
        if hdr.messagetype == DsdvMessageTypes.fullDumpUpdate:
            self.process_full_dump_update_message(payload, hdr.messagefrom)
        elif hdr.messagetype == DsdvMessageTypes.incrementalUpdate:
            self.process_incremental_update_message(payload, hdr.messagefrom)
        #---------------

        #------Data messages----
        elif hdr.messagetype == DataMessageTypes.appData:
            # If destination of the message is myself, send to higher layer
            if hdr.messageto == self.componentinstancenumber:
                self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))
            # If destination of the message is not myself, work as a router: Decide a path and resend the message
            else:
                newMessage = self.RouteAppLayerMessage(hdr.messageto, msg.payload)
                self.send_down(Event(self, EventTypes.MFRT, newMessage))

    # This function handles the messages coming from upper layer
    # Function overrides super class function
    #
    def on_message_from_top(self, eventobj: Event):
        msg = eventobj.eventcontent
        hdr = msg.header
        newMessage = self.RouteAppLayerMessage(hdr.messageto, msg.payload)
        if(newMessage != -1):
            self.send_down(Event(self, EventTypes.MFRT, newMessage))

    # This function routes a message according to self routing table.
    # The function takes destination address and payload information and regenerates the message
    # @param destination is the destination address of the message
    # @param payload is the payload part of the message
    #
    def RouteAppLayerMessage(self, destination, payload):
        entryIndex = self.search_for_entry(destination)

        if entryIndex == -1:  # No match in self.routing table. Drop the packet
            print("I am: " + self.unique_name() + " Message is dropped")
            print(payload)
            return -1
        else:
            nexthop = self.routingTable[entryIndex][RoutingTableColumn.nextHop.value]
            header = GenericMessageHeader(DataMessageTypes.appData, self.componentinstancenumber, destination,
                                          nexthop)

            message = GenericMessage(header, payload)
            return message














import sys
import queue
import random
from enum import Enum
from datetime import datetime as dt

from ahc.LinkLayers.GenericLinkLayer import LinkLayer
from ahc.Routing.AllSeeingEyeNetworkLayer import AllSeingEyeNetworkLayer
from ahc.Ahc import (ComponentModel, Event, ConnectorTypes, ComponentRegistry, GenericMessagePayload, GenericMessageHeader,
                 GenericMessage, EventTypes)

WAVE_DEBUG = False

registry = ComponentRegistry()

# define your own message types
class ApplicationLayerMessageType(Enum):
    BASIC = "basic"
    CONTROL = "control" # NOTE: Means the acknowledgement message!
    WAVE = "wave"

class SFWaveMessageType(Enum):
    REQUEST = "req"
    RESPONSE = "resp"

    def __str__(self):
        return self.name

class SFWaveResponse(Enum):
    FINISHED = "fin"
    ACTIVE = "act"

    def __str__(self):
        return self.name

class SFWaveMessagePayload:
    tag: int
    type: SFWaveMessageType
    response: SFWaveResponse

    def __init__(self, tag: int, typ: SFWaveMessageType, resp: SFWaveResponse = None) -> None:
        self.tag = tag
        self.response = resp
        self.type = typ

class SFAHCNodeSimulationStatus(Enum):
    ACTIVE = "active"
    PASSIVE = "passive"
    OUT_OF_CLOCK = "ooc"
    OUT_OF_TREE = "oot"

    def __str__(self):
        return self.name

# define your own message header structure
class ApplicationLayerMessageHeader(GenericMessageHeader):
    pass

# define your own message payload structure
class ApplicationLayerMessagePayload(GenericMessagePayload):
    pass

class ShavitFrancezApplicationLayerComponent(ComponentModel):
    def __init__(self, componentname, componentinstancenumber, context):
        super().__init__(componentname, componentinstancenumber, context=context)

        self.context = context
        # self.eventhandlers[ApplicationLayerMessageType.BASIC] = self.on_basic_message
        # self.eventhandlers[ApplicationLayerMessageType.CONTROL] = self.on_control_message

        self.basic_message_queue = queue.Queue(maxsize=-1)
        self.control_message_queue = queue.Queue(maxsize=-1)
        self.simulation_state = SFAHCNodeSimulationStatus.PASSIVE

        self.sleep_ms_per_tick = context["ms_per_tick"]
        self.simulation_ticks_total = context["simulation_ticks"]
        self.alive_for_next_ticks = context["initial_liveness"][componentinstancenumber]
        self.communication_on_active_prob = context["communication_on_active_prob"]
        self.min_activeness_after_receive = context["min_activeness_after_receive"]
        self.max_activeness_after_receive = context["max_activeness_after_receive"]
        self.package_process_per_tick = context["node_package_process_per_tick"]
        self.die_passiveness_threshold = context["passiveness_death_thresh"]

        if context["hard_stop_on_tick"] is None:
            self.hard_stop_on_tick = None
        else:
            self.hard_stop_on_tick = context["hard_stop_on_tick"][self.componentinstancenumber]

        self.__tick_n = 0
        self._passive_counter = 0

        self._child_counter = 0
        self._parent_node = None
        self._in_tree = False
        self._children = []
        
        self._cms = 0   # control messages sent
        self._wms = 0   # wave messages sent

        self._i_am_root = context["network"].root == self.componentinstancenumber

        if context["only_root_alive_initially"]:
            if self._i_am_root:
                self.alive_for_next_ticks = 20 # this may change though...
                self._in_tree = True
            else:
                self.alive_for_next_ticks = 0
        elif self._i_am_root and self.alive_for_next_ticks == 0:
            self.alive_for_next_ticks = 20 # this may change though...

        if self.alive_for_next_ticks > 0:
            self.simulation_state = SFAHCNodeSimulationStatus.ACTIVE

        self.__exited_from_tree = False
        self.my_wave_bucket = []
        self.announce_on_next_tick = False

    def prepare_application_layer_message(self, message_type: ApplicationLayerMessageType, destination_node_id: int, payload: object) -> GenericMessage:
        hdr = ApplicationLayerMessageHeader(message_type, self.componentinstancenumber, destination_node_id)
        payload = ApplicationLayerMessagePayload(payload)
        
        return GenericMessage(hdr, payload)

    def send_random_basic_message(self, to: int) -> None:
        self._child_counter += 1
        self._children.append(to)
        self.send_down(Event(self, EventTypes.MFRT, self.prepare_application_layer_message(ApplicationLayerMessageType.BASIC, to, str(dt.now().timestamp()))))

    def send_ack_control_message(self, to: int, is_dead: bool) -> None:
        # print(f"send_ack_control_message: N-{self.componentinstancenumber} ==> N-{to} ({self._parent_node}) : {'DEAD' if is_dead else 'PKG_RESP'}")
        self.send_down(Event(self, EventTypes.MFRT, self.prepare_application_layer_message(ApplicationLayerMessageType.CONTROL, to, str(dt.now().timestamp()))))
        self._cms += 1

    def on_init(self, eventobj: Event):
        # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
        pass

    def on_message_from_bottom(self, eventobj: Event):
        applmessage = eventobj.eventcontent
        hdr = applmessage.header

        # print(f"Node-{self.componentinstancenumber}: Node-{hdr.messagefrom} has sent {hdr.messagetype} message (payload: {applmessage.payload})")

        if hdr.messagetype == ApplicationLayerMessageType.BASIC:
            self.basic_message_queue.put_nowait(applmessage)

            if self._in_tree:
                self.send_ack_control_message(hdr.messagefrom, False)
            else:
                self._parent_node = hdr.messagefrom
                self._in_tree = True

                if self.componentinstancenumber not in self.context["alive_nodes"]:
                    self.context["alive_nodes"].append(self.componentinstancenumber)
        elif hdr.messagetype == ApplicationLayerMessageType.CONTROL:
            # self.control_message_queue.put_nowait(applmessage)

            try:
                self._children.remove(hdr.messagefrom)
                self._child_counter -= 1
            except ValueError as e:
                # print(f"\n\n\n{self.componentinstancenumber}: {e} {hdr.messagefrom} {self._children}\n\n\n")
                pass
        elif hdr.messagetype == ApplicationLayerMessageType.WAVE:
            if applmessage.payload.messagepayload.tag == self.componentinstancenumber:
                self.my_wave_bucket.append(applmessage.payload.messagepayload.response)

                if WAVE_DEBUG:
                    print(f"  WAVE << N-{self.componentinstancenumber} << N-{hdr.messagefrom} ({applmessage.payload.messagepayload.response})")

                if len(self.my_wave_bucket) == (len(self.context["network"].G.nodes()) - 1):
                    if WAVE_DEBUG:
                        print(f"  ::: WAVE END >> N-{self.componentinstancenumber}: {' '.join(['A' if x == SFWaveResponse.ACTIVE else ('F' if x == SFWaveResponse.FINISHED else 'XXX') for x in self.my_wave_bucket])}")

                    if SFWaveResponse.ACTIVE not in self.my_wave_bucket:
                        print(f"  ::: WAVE END >> N-{self.componentinstancenumber}: ANNOUNCE!")
                        self.announce_on_next_tick = True
                    else:
                        print(f"  ::: WAVE END >> N-{self.componentinstancenumber}: NOT DONE YET!")
                        self.my_wave_bucket = []
            else:
                if self.__exited_from_tree:
                    self.send_down(Event(self, EventTypes.MFRT, self.prepare_application_layer_message(ApplicationLayerMessageType.WAVE, applmessage.payload.messagepayload.tag, SFWaveMessagePayload(applmessage.payload.messagepayload.tag, SFWaveMessageType.RESPONSE, SFWaveResponse.FINISHED))))
                    if WAVE_DEBUG:
                        print(f"  WAVE >> N-{self.componentinstancenumber} >> N-{applmessage.payload.messagepayload.tag} >> FINISHED")
                else:
                    self.send_down(Event(self, EventTypes.MFRT, self.prepare_application_layer_message(ApplicationLayerMessageType.WAVE, applmessage.payload.messagepayload.tag, SFWaveMessagePayload(applmessage.payload.messagepayload.tag, SFWaveMessageType.RESPONSE, SFWaveResponse.ACTIVE))))
                    if WAVE_DEBUG:
                        print(f"  WAVE >> N-{self.componentinstancenumber} >> N-{applmessage.payload.messagepayload.tag} >> ACTIVE")

                self._wms += 1
        else:
            print(f"\n!!! N-{self.componentinstancenumber}: GOT MSG UNIDENT: {hdr.messagetype}, FROM {hdr.messagefrom}")
            sys.exit(2)

    def call_wave(self):
        print(f"  > CALL WAVE @@ N-{self.componentinstancenumber}")

        to_nodes = [i for i in range(len(self.context["network"].G.nodes())) if i != self.componentinstancenumber]

        for to in to_nodes:
            self.send_down(Event(self, EventTypes.MFRT, self.prepare_application_layer_message(ApplicationLayerMessageType.WAVE, to, SFWaveMessagePayload(self.componentinstancenumber, SFWaveMessageType.REQUEST))))
            self._wms += 1

    def exit_tree(self):
        # Exit from the tree

        if not self.__exited_from_tree:
            if self._in_tree:
                self._in_tree = False
                if self._parent_node is not None:
                    self.send_ack_control_message(self._parent_node, True)
            
            self._parent_node = None

            if self.componentinstancenumber in self.context["alive_nodes"]:
                self.context["alive_nodes"].remove(self.componentinstancenumber)

            self.__exited_from_tree = True
            print(f"  > EXIT @@ N-{self.componentinstancenumber}")

            self.call_wave()

    def simulation_tick(self):
        next_state = None
        got_packages_from = None
        to_friend = None

        _upd_children = []

        for c in self._children:
            if c in self.context["alive_nodes"]:
                _upd_children.append(c)

        self._children = _upd_children
        self._child_counter = len(self._children)

        if self.announce_on_next_tick:
            __cms =  self._cms
            __wms = self._wms
            self._cms = 0
            self._wms = 0

            return None, None, __cms, __wms

        if self.simulation_state == SFAHCNodeSimulationStatus.OUT_OF_TREE:
            next_state = SFAHCNodeSimulationStatus.OUT_OF_TREE
            # print(f"   ==> N-{self.componentinstancenumber}: OOT")            #NOTE: DEV
            # print(f"   ==> N-{self.componentinstancenumber}: OUT OF TREE")    
        elif self.__tick_n >= self.simulation_ticks_total:
            self.exit_tree()
            next_state = SFAHCNodeSimulationStatus.OUT_OF_TREE
            # print(f"   ==> N-{self.componentinstancenumber}: OOC DEAD")       #NOTE: DEV
        elif self._passive_counter >= self.die_passiveness_threshold:           # NOTE: initiator can also die from passiveness in Shavit-Francez.
            self.exit_tree()
            next_state = SFAHCNodeSimulationStatus.OUT_OF_TREE
            # print(f"   ==> N-{self.componentinstancenumber}: PASSIVE DIE")    #NOTE: DEV
        elif self.__tick_n >= self.hard_stop_on_tick:                           # NOTE: initiator can also hard stop in Shavit-Francez.
            self.exit_tree()
            next_state = SFAHCNodeSimulationStatus.OUT_OF_TREE
            # print(f"   ==> N-{self.componentinstancenumber}: HARD STOP")      #NOTE: DEV
        else:
            if self.simulation_state == SFAHCNodeSimulationStatus.OUT_OF_CLOCK:
                next_state = SFAHCNodeSimulationStatus.OUT_OF_CLOCK
            elif self.simulation_state == SFAHCNodeSimulationStatus.PASSIVE:
                if self.basic_message_queue.empty():
                    if self._in_tree and self._child_counter == 0:
                            self.exit_tree()
                            next_state = SFAHCNodeSimulationStatus.PASSIVE
                            print(f"   ==> N-{self.componentinstancenumber}: OUT OF TREE / PASSIVE")
                    else:
                        # no incoming package, still passive.
                        next_state = SFAHCNodeSimulationStatus.PASSIVE
                else:
                    got_packages_from = []

                    for _ in range(self.package_process_per_tick):
                        try:
                            package = self.basic_message_queue.get_nowait()
                            # print(f"+P+ N-{self.componentinstancenumber} <==BASIC== N-{package.header.messagefrom} ({package.payload.messagepayload})")
                            got_packages_from.append(package.header.messagefrom)
                        except queue.Empty:
                            break

                    self.alive_for_next_ticks = random.randint(self.min_activeness_after_receive, self.max_activeness_after_receive)
                    next_state = SFAHCNodeSimulationStatus.ACTIVE
            elif self.simulation_state == SFAHCNodeSimulationStatus.ACTIVE:
                got_packages_from = []

                for _ in range(self.package_process_per_tick):
                    try:
                        package = self.basic_message_queue.get_nowait()
                        # print(f"+A+ N-{self.componentinstancenumber} <==BASIC== N-{package.header.messagefrom} ({package.payload.messagepayload})")
                        got_packages_from.append(package.header.messagefrom)
                    except queue.Empty:
                        break

                if random.random() <= self.communication_on_active_prob:
                    # send package to a random friend..

                    _alive_ones = [n for n in self.context["alive_nodes"] if n != self.componentinstancenumber]

                    if len(_alive_ones) == 0: # everyone is dead!!!
                        # print(f"  **ROOT** N-{self.componentinstancenumber}: Eveyone is dead!!!")
                        # return None, None       # time to go!
                        to_friend = None
                    else:
                        to_friend = random.choice(_alive_ones)
                        self.send_random_basic_message(to_friend)

                self.alive_for_next_ticks -= 1

                if self.alive_for_next_ticks == 0:
                    if len(got_packages_from) > 0:
                        # got a package, this means immeiate activeness from passive!
                        next_state = SFAHCNodeSimulationStatus.ACTIVE
                        self.alive_for_next_ticks = random.randint(self.min_activeness_after_receive, self.max_activeness_after_receive)
                    else:
                        next_state = SFAHCNodeSimulationStatus.PASSIVE
                else:
                    next_state = SFAHCNodeSimulationStatus.ACTIVE

        assert next_state is not None

        # ST: state
        # NS: next state
        # GPF: got packages from
        # SPF: sent package to friend
        # ANT: alive next ticks
        # P2P: packages to process
        # print(f"   {'ROOT' if self._i_am_root else '==>'} N-{self.componentinstancenumber}: P: {self._parent_node}, CC: ({self._child_counter}) {self._children}, ST: {self.simulation_state}, NS: {next_state}, GPF: {got_packages_from}, SPF: {to_friend}, ANT: {self.alive_for_next_ticks}, P2P: {self.basic_message_queue.qsize()}")

        if self._i_am_root:
            print(f"   {'INIT' if self._i_am_root else '==>'} N-{self.componentinstancenumber}: P: {self._parent_node}, CC: ({self._child_counter}) {self._children}, ST: {self.simulation_state}, NS: {next_state}, GPF: {got_packages_from}, SPF: {to_friend}, ANT: {self.alive_for_next_ticks}, P2P: {self.basic_message_queue.qsize()}")

        # time.sleep(self.sleep_ms_per_tick / 1000)
        self.__tick_n += 1
        self.simulation_state = next_state

        if self.simulation_state == SFAHCNodeSimulationStatus.PASSIVE:
            self._passive_counter += 1
        elif self.simulation_state == SFAHCNodeSimulationStatus.ACTIVE:
            self._passive_counter = 0

        __cms =  self._cms
        __wms = self._wms
        self._cms = 0
        self._wms = 0

        return next_state, to_friend, __cms, __wms

class ShavitFrancezAdHocNode(ComponentModel):
    def __init__(self, componentname, componentid, context):
        self.context = context
        # SUBCOMPONENTS
        self.appllayer = ShavitFrancezApplicationLayerComponent("ApplicationLayer", componentid, context=self.context)
        self.netlayer = AllSeingEyeNetworkLayer("NetworkLayer", componentid)
        self.linklayer = LinkLayer("LinkLayer", componentid)
        # self.failuredetect = GenericFailureDetector("FailureDetector", componentid)

        # CONNECTIONS AMONG SUBCOMPONENTS
        self.appllayer.connect_me_to_component(ConnectorTypes.DOWN, self.netlayer)
        # self.failuredetect.connectMeToComponent(PortNames.DOWN, self.netlayer)
        self.netlayer.connect_me_to_component(ConnectorTypes.UP, self.appllayer)
        # self.netlayer.connectMeToComponent(PortNames.UP, self.failuredetect)
        self.netlayer.connect_me_to_component(ConnectorTypes.DOWN, self.linklayer)
        self.linklayer.connect_me_to_component(ConnectorTypes.UP, self.netlayer)

        # Connect the bottom component to the composite component....
        self.linklayer.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.linklayer)

        super().__init__(componentname, componentid, context=self.context)

    def on_init(self, eventobj: Event):
        # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
        pass

    def on_message_from_top(self, eventobj: Event):
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))

    def simulation_tick(self):
        return self.appllayer.simulation_tick()

    @property
    def waiting_packages_on_queue(self):
        if self.appllayer.simulation_state == SFAHCNodeSimulationStatus.OUT_OF_CLOCK or self.appllayer.simulation_state == SFAHCNodeSimulationStatus.OUT_OF_TREE:
            return 0

        return self.appllayer.basic_message_queue.qsize()

    @property
    def parent_node(self):
        return self.appllayer._parent_node


from argparse import ArgumentParser
from simulator_sf import run_shavit_francez_simulation
from simulator_ds import run_dijkstra_scholten_simulation

algorithm_handlers = {
    "djikstra-scholten": run_dijkstra_scholten_simulation,
    "shavit-francez": run_shavit_francez_simulation,
}

    ## IMPORTANT NOTE
    # possibe evaluation metrics
    # num(control packages) / num(total packages) => 
    # tick(algo done) - tick(termination) => ne kadar erken detect etti

parser = ArgumentParser()
parser.add_argument("algorithm", type=str, choices=list(algorithm_handlers.keys()))
parser.add_argument("simulation_ticks", type=int)
parser.add_argument("ms_per_tick", type=int)
parser.add_argument("--node_min_activeness_after_receive", type=int, default=3) # paket aldiktan sonra min bu kadar aktif kal
parser.add_argument("--node_max_activeness_after_receive", type=int, default=5) # paket aldiktan sonra max bu kadar aktif kal
parser.add_argument("--node_activeness_communication_prob", type=float, default=0.5) # alive iken baska nodelara paket gonderme olasiligi
parser.add_argument("--node_initial_activeness_prob", type=float, default=0.5)
parser.add_argument("--node_package_process_per_tick", type=int, default=5)

parser.add_argument("--only_root_alive_initially", action="store_true", default=False)

parser.add_argument("--run_until_termination", action="store_true", default=False)
parser.add_argument("--exit_on_termination", action="store_true", default=False)
parser.add_argument("--wait_ticks_after_termination", type=int, default=0)

parser.add_argument("--passiveness_death_thresh", type=int, default=20)
parser.add_argument("--hard_stop_nodes", action="store_true", default=False)
parser.add_argument("--hard_stop_min_tick", type=int, default=50)
parser.add_argument("--hard_stop_max_tick", type=int, default=300)
parser.add_argument("--hard_stop_prob", type=float, default=0.5)

parser.add_argument("--no_realtime_plot", action="store_true", default=False)
parser.add_argument("--save_tick_plots", action="store_true", default=False)
parser.add_argument("--tick_plots_save_dir", type=str, default="simdump")
parser.add_argument("--generate_gif", action="store_true", default=False)

sp = parser.add_subparsers()

erg_parser = sp.add_parser("erg")
erg_parser.add_argument("node_count", type=int)
erg_parser.add_argument("--node_connectivity", type=float, default=0.5)
erg_parser.set_defaults(network_type="erg")

grid_parser = sp.add_parser("grid")
grid_parser.add_argument("node_count_on_edge", type=int)
grid_parser.set_defaults(network_type="grid")

star_parser = sp.add_parser("star")
star_parser.add_argument("slave_count", type=int)
star_parser.add_argument("--master_is_root", type=bool, default=True)
star_parser.set_defaults(network_type="star")

if __name__ == "__main__":
    args = parser.parse_args()
    print(f"[+] Network type: {args.network_type}")

    if args.generate_gif:
        args.save_tick_plots = True

    algorithm_handlers[args.algorithm](args)


from ahc.Ahc import \
    ComponentModel, \
    ComponentRegistry, \
    EventTypes, \
    GenericMessage, \
    GenericMessageHeader, \
    Lock, \
    Event, \
    Thread, \
    Topology
from ahc.Routing.GSR.GSRQueueElement import GSQQueueElement
from ahc.Routing.GSR.GSRExperimentDataCollector import GSRExperimentCollector
from ahc.Routing.GSR.Constants import GSR_COORDINATOR_NAME, \
    GSR_ROUTER_NAME, \
    UPDATE_MESSAGE_TYPE, \
    TERMINATE_ROUTING_MESSAGE_TYPE, \
    ROUTING_COMPLETED_MESSAGE_TYPE, \
    ROUTER_UPDATE_PERIOD_SECS, \
    ENABLE_NETWORK_LEVEL_LOGGING
import time


class RoutingGSRComponent(ComponentModel):
    def __init__(self, component_name, component_id):
        super(RoutingGSRComponent, self).__init__(component_name, component_id)
        self.start_time = 0
        self.n_control_messages = 0
        self.terminated = False
        self.routing_completed = False
        self.neighbors = []
        self.n_nodes = 1
        self.distances = {}
        self.link_states = {}
        self.sequence_numbers = {}
        self.next_hop = {}
        self.message_queue = []
        self.queue_lock = Lock()

    def on_init(self, eventobj: Event):
        super(RoutingGSRComponent, self).on_init(eventobj)
        self.neighbors = Topology().get_neighbors(self.componentinstancenumber)

        self.n_nodes = self.componentinstancenumber
        for element in ComponentRegistry().components:
            if RoutingGSRComponent.__name__ in element:
                component_id = int(element.split(RoutingGSRComponent.__name__)[1])
                self.n_nodes = max(self.n_nodes, component_id + 1)

        self.distances = {i: -1 for i in range(self.n_nodes)}
        empty_link_state = {i: -1 for i in range(self.n_nodes)}
        self.link_states = {i: empty_link_state.copy() for i in range(self.n_nodes)}
        self.sequence_numbers = {i: -1 for i in range(self.n_nodes)}
        self.next_hop = {i: -1 for i in range(self.n_nodes)}

        self.distances[self.componentinstancenumber] = 0
        self.link_states[self.componentinstancenumber][self.componentinstancenumber] = 0
        self.sequence_numbers[self.componentinstancenumber] = 0
        self.next_hop[self.componentinstancenumber] = self.componentinstancenumber

        self.start_time = time.time()
        thread = Thread(target=self.job, args=[14, 23])
        thread.start()

    def on_message_from_bottom(self, eventobj: Event):
        message_header = eventobj.eventcontent.header
        message_destination = message_header.messageto.split("-")[0]
        if message_destination == GSR_ROUTER_NAME:
            message_source_id = message_header.messagefrom.split("-")[1]
            message_type = message_header.messagetype
            content = eventobj.eventcontent.payload

            if message_type == UPDATE_MESSAGE_TYPE:
                self.queue_lock.acquire()
                self.message_queue.append(GSQQueueElement(message_source_id, content))
                self.n_control_messages += 1
                self.queue_lock.release()

    def on_message_from_peer(self, eventobj: Event):
        message_header = eventobj.eventcontent.header
        message_destination = message_header.messageto.split("-")[0]
        message_type = message_header.messagetype
        if message_destination == GSR_ROUTER_NAME:
            if message_type == TERMINATE_ROUTING_MESSAGE_TYPE:
                self.terminated = True

    def job(self, *arg):
        while not self.terminated:
            self.queue_lock.acquire()
            for pkt in self.message_queue:
                self.pkt_process(pkt)
            self.message_queue = []
            self.queue_lock.release()
            self.find_shortest_paths()
            self.broadcast_routing_update()
            if self.routing_completed:
                self.report_route()
            time.sleep(ROUTER_UPDATE_PERIOD_SECS)

    def pkt_process(self, pkt: GSQQueueElement):
        self.link_states[self.componentinstancenumber][pkt.source_id] = pkt.transfer_duration
        for i in range(self.n_nodes):
            if i != self.componentinstancenumber and pkt.sequence_numbers[i] > self.sequence_numbers[i]:
                self.sequence_numbers[i] = pkt.sequence_numbers[i]
                self.link_states[i] = pkt.link_states[i].copy()

    def broadcast_routing_update(self):
        self.sequence_numbers[self.componentinstancenumber] += 1
        message_from = GSR_ROUTER_NAME + "-" + str(self.componentinstancenumber)
        payload = {
            "link_states": self.link_states,
            "sequence_numbers": self.sequence_numbers,
            "timestamp": time.time() * 1000
        }

        for neighbor_id in self.neighbors:
            message_to = GSR_ROUTER_NAME + "-" + str(neighbor_id)
            interface_id = str(self.componentinstancenumber) + "-" + str(neighbor_id)
            message_header = GenericMessageHeader(
                UPDATE_MESSAGE_TYPE,
                message_from,
                message_to,
                interfaceid=interface_id
            )
            message = GenericMessage(message_header, payload)
            event = Event(self, EventTypes.MFRT, message)
            self.send_down(event)

    def report_route(self):
        message_from = GSR_ROUTER_NAME + "-" + str(self.componentinstancenumber)
        message_to = GSR_COORDINATOR_NAME + "-" + str(self.componentinstancenumber)
        message_header = GenericMessageHeader(ROUTING_COMPLETED_MESSAGE_TYPE, message_from, message_to)
        payload = {"routing_table": self.next_hop}
        message = GenericMessage(message_header, payload)
        event = Event(self, EventTypes.MFRP, message)
        if ENABLE_NETWORK_LEVEL_LOGGING:
            print("SENDING [" + message_from + " -> " + message_to + "]: " + str(payload))
        self.send_peer(event)
        if self.componentinstancenumber == 0:
            if GSRExperimentCollector().first_routing_completion == 0:
                time_passed = time.time() - self.start_time
                GSRExperimentCollector().first_routing_completion = time_passed
                GSRExperimentCollector().n_control_messages = self.n_control_messages

    def find_shortest_paths(self):
        processed_nodes = [self.componentinstancenumber]
        self.distances[self.componentinstancenumber] = 0
        for i in range(self.n_nodes):
            if self.link_states[self.componentinstancenumber][i] >= 0:
                self.distances[i] = self.link_states[self.componentinstancenumber][i]
                self.next_hop[i] = i
            else:
                self.distances[i] = -1
                self.next_hop[i] = -1

        while len(processed_nodes) < self.n_nodes:
            node_k = 0
            node_l = 0
            min_w = -1
            for k in range(self.n_nodes):
                if k in processed_nodes:
                    continue
                for l in processed_nodes:
                    if self.link_states[l][k] == -1:
                        continue
                    w = self.distances[l] + self.link_states[l][k]
                    if min_w == -1 or w < min_w:
                        node_k = k
                        node_l = l
                        min_w = w
            if min_w == -1:
                break
            processed_nodes.append(node_k)
            self.distances[node_k] = min_w
            self.next_hop[node_k] = self.next_hop[node_l] if node_l != self.componentinstancenumber else node_k

        if ENABLE_NETWORK_LEVEL_LOGGING:
            print("Next hops for node " + str(self.componentinstancenumber) + ": " + str(self.next_hop))
        self.routing_completed = len(processed_nodes) == self.n_nodes



from ahc.Ahc import Topology
from Channels.Channels import  P2PFIFOPerfectChannel
import networkx as nx
from Routing.FredericksonAlgorithmSimple.RoutingExample.MachineLearningNodeComponent import MachineLearningNode

edges = [(0, 1, {"weight": 1}), (0, 2, {"weight": 1}), (1, 3, {"weight": 1}), (2, 4, {"weight": 1}), (4, 5, {"weight": 1}),
         (3, 5, {"weight": 1})]

# undirected graph
graph = nx.Graph()
graph.add_edges_from(edges)

topology = Topology()
topology.construct_from_graph(graph, MachineLearningNode, P2PFIFOPerfectChannel)
# process1 = MachineLearningNode("MachineLearningNode", 0)
# ComponentRegistry().init()

topology.start()
while True: pass




import time
from enum import Enum
from ahc.Ahc import ComponentModel, Event, ConnectorTypes, Topology
from ahc.Ahc import ComponentRegistry
from ahc.Ahc import GenericMessagePayload, GenericMessageHeader, GenericMessage, EventTypes

registry = ComponentRegistry()
topo = Topology()

rerr = 1


class AODV_ABRMessageTypes(Enum):
    RREQ = "RREQ"
    RREP = "RREP"
    RERR = "RERR"
    BRRQ = "BRRQ"
    BRRP = "BRRP"
    OVRH = "OVRH"
    DATA = 'DATA'


# # define your own message header structure
class AODV_ABRMessageHeader(GenericMessageHeader):
    def __init__(self, *args):
        super().__init__(*args[:6])
        self.rreq_id = args[6]
        self.hop_count = args[7]
        self.int_sender = args[8]


# define your own message payload structure
class AODV_ABRMessagePayload(GenericMessagePayload):
    def __init__(self, *args):
        super().__init__(*args[:1])
        self.toNode = args[1]


class AODV_ABRComponent(ComponentModel):

    def __init__(self, componentname, componentinstancenumber):
        print(f"Initializing {componentname}.{componentinstancenumber}")
        super().__init__(componentname, componentinstancenumber)
        neighbour_list = topo.get_neighbors(self.componentinstancenumber)
        self.NeighbourList = neighbour_list
        self.RoutingTable = dict()
        self.AlternateRouteTable = dict()
        self.rreq_list = set()
        self.BBRP_Responses = dict()

    def on_message_from_bottom(self, eventobj: Event):
        msg = eventobj.eventcontent
        hdr = msg.header
        py = msg.payload
        message_source = hdr.messagefrom

        if hdr.messagetype == AODV_ABRMessageTypes.RREQ:
            self.updateRoutingTableNeighbours()

            if hdr.rreq_id in self.rreq_list:
                # print('Ignoring duplicate RREQ message with rreq id ', hdr.rreq_id)
                return
            else:
                self.rreq_list.add(hdr.rreq_id)
                if message_source in self.RoutingTable.keys():  # update the route table if the entries are not fresh
                    route = self.RoutingTable[message_source]
                    if hdr.sequencenumber > int(route['Seq_No']):
                        route['Seq_No'] = hdr.sequencenumber
                    elif hdr.sequencenumber == int(route['Seq_No']) and hdr.hop_count < route['Hop_Count']:
                        route['Hop_Count'] = hdr.hop_count
                        route['Next_Hop'] = message_source
                else:
                    if hdr.messagefrom != self.componentinstancenumber:  # if the source is not in routing table, create an entry
                        self.RoutingTable[message_source] = {'Dest': str(message_source),
                                                            'Next_Hop': str(hdr.int_sender),
                                                            'Seq_No':  str(hdr.sequencenumber),
                                                            'Hop_Count': str(hdr.hop_count + 1)}
                    self.show_routing_table()

                if hdr.messageto == self.componentinstancenumber:  # if we are the destination, send RREP
                    self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.RREP,
                                                                                     message_source,
                                                                                     hdr.messageto,
                                                                                     hdr.int_sender,
                                                                                     0, hdr.sequencenumber + 1, -1, 0, self.componentinstancenumber)))
                    return

                if hdr.messageto in self.RoutingTable.keys():  # if the dest is in Routing Table, check the freshness
                    if hdr.sequencenumber <= int(self.RoutingTable[hdr.messageto]['Seq_No']):
                        self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.RREQ,
                                                                                          message_source,
                                                                                          hdr.messageto, int(self.RoutingTable[hdr.messageto]['Next_Hop']),
                                                                                          0, hdr.sequencenumber, -1, hdr.hop_count+1, self.componentinstancenumber)))

                        return
                    else:
                        pass

                else:
                    for i in self.NeighbourList: # if the dest is not in routing table, broadcast
                        self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.RREQ,
                                                                                                 message_source,
                                                                                                 hdr.messageto,
                                                                                                 int(i),
                                                                                                 0, hdr.sequencenumber, hdr.rreq_id,
                                                                                                 hdr.hop_count, self.componentinstancenumber)))

        if hdr.messagetype == AODV_ABRMessageTypes.RREP:
            self.updateRoutingTableNeighbours()

            if hdr.messagefrom == self.componentinstancenumber: # if source node receives the RREP, routing finishes
                self.RoutingTable[hdr.messageto] = {'Dest': str(hdr.messageto),
                                                      'Next_Hop': str(hdr.int_sender),
                                                      'Seq_No':  str(hdr.sequencenumber),
                                                      'Hop_Count': str(hdr.hop_count + 1)}
                #self.show_routing_table()
                print('End of Routing')
                return
            else: # forward the RREP, update the Route Table for the dest, also send the OVERHEAR info to the neighbours
                if hdr.messageto not in self.RoutingTable:
                    self.RoutingTable[hdr.messageto] = {'Dest': str(hdr.messageto),
                                                         'Next_Hop': str(hdr.int_sender),
                                                         'Seq_No': str(hdr.sequencenumber),
                                                         'Hop_Count': str(hdr.hop_count + 1)}
                    #self.show_routing_table()

                hopCount = int(self.RoutingTable[hdr.messageto]['Hop_Count'])
                self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.RREP,
                                                                                 message_source,
                                                                                 hdr.messageto,
                                                                                 int(self.RoutingTable[message_source]['Next_Hop']),
                                                                                 0, hdr.sequencenumber, -1, hopCount,
                                                                                 self.componentinstancenumber)))

                for i in self.NeighbourList:
                    if i != int(self.RoutingTable[hdr.messagefrom]['Next_Hop']) and i != int(hdr.int_sender):
                        hopCount = int(self.RoutingTable[hdr.messageto]['Hop_Count'])
                        self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.OVRH,
                                                                                         message_source,
                                                                                         hdr.messageto,
                                                                                         int(i),
                                                                                         0, hdr.sequencenumber, -1, hopCount,
                                                                                         self.componentinstancenumber)))
        if hdr.messagetype == AODV_ABRMessageTypes.OVRH:
            self.updateRoutingTableNeighbours()
            if hdr.messageto in self.AlternateRouteTable.keys(): # if dest is in AlternateRouteTable, check freshness
                route = self.AlternateRouteTable[hdr.messageto]
                if hdr.sequencenumber > int(route['Seq_No']):
                    route['Seq_No'] = hdr.sequencenumber
                elif hdr.sequencenumber == int(route['Seq_No']) and hdr.hop_count < int(route['Hop_Count']):
                    route['Hop_Count'] = hdr.hop_count
                    route['Next_Hop'] = message_source
                else:
                    print('Ignoring duplicate OVERHEAR messages')
                    pass
                self.show_alternate_route_table()
            else:
                if hdr.messageto not in self.RoutingTable: # if the destination is not in Alternate Route table AND in Routing Table, create entry
                    self.AlternateRouteTable[hdr.messageto] = {'Dest': str(hdr.messageto),
                                                             'Next_Hop': str(hdr.int_sender),
                                                             'Seq_No': str(hdr.sequencenumber),
                                                             'Hop_Count': str(hdr.hop_count + 1)}

                    self.show_alternate_route_table()
                else:
                    #print('if the dest is in Routing Table, no new entry in Alternate Table')
                    pass

        if hdr.messagetype == AODV_ABRMessageTypes.BRRQ:
            self.updateRoutingTableNeighbours()
            global rerr
            if hdr.messagefrom == self.componentinstancenumber: #sending BRRQ to neighbours
                for i in self.NeighbourList:
                    if i != py.toNode:
                        self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.BRRQ,
                                                                                         message_source,
                                                                                         hdr.messageto,
                                                                                         int(i),
                                                                                         0, hdr.sequencenumber, hdr.rreq_id,
                                                                                         hdr.hop_count,
                                                                                         self.componentinstancenumber, py.messagepayload, py.toNode)))

                time.sleep(5) #20
                if rerr == 1:
                    print('********rerr is 1 ***********')
                    self.show_routing_table()
                    self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.RERR,
                                                                                     py.messagepayload,
                                                                                     hdr.messageto,
                                                                                     int(self.RoutingTable[py.messagepayload]['Next_Hop']),
                                                                                     0, hdr.sequencenumber, hdr.rreq_id,
                                                                                     hdr.hop_count,
                                                                                     self.componentinstancenumber, py.messagepayload, py.toNode)))

            else:
                #print('Inside brrq for sending back brrp, I am node ', self.componentinstancenumber)
                self.show_alternate_route_table()
                if hdr.messageto in self.AlternateRouteTable.keys() and int(self.AlternateRouteTable[hdr.messageto]['Next_Hop']) != int(hdr.int_sender):
                    rerr = 0
                    #print(self.componentinstancenumber, ' have dest entry in its alternate route table')
                    route = int(self.AlternateRouteTable[hdr.messageto]['Hop_Count']) ##

                    self.RoutingTable[hdr.messageto] = {'Dest': str(hdr.messageto),
                                                        'Next_Hop': self.AlternateRouteTable[hdr.messageto]['Next_Hop'],
                                                        'Seq_No': str(hdr.sequencenumber),
                                                        'Hop_Count': str(route)}
                    self.show_routing_table()

                    self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.BRRP,
                                                                                     message_source,
                                                                                     hdr.messageto,
                                                                                     message_source,
                                                                                     0, hdr.sequencenumber, hdr.rreq_id,
                                                                                     route,
                                                                                     self.componentinstancenumber, py.messagepayload, py.toNode)))



                elif hdr.messageto in self.RoutingTable.keys() and int(self.RoutingTable[hdr.messageto]['Next_Hop']) != int(hdr.int_sender): ################
                    rerr = 0
                    #print(self.componentinstancenumber, ' have dest entry in its route table')
                    time.sleep(10)
                    self.show_routing_table()
                    route = self.RoutingTable[hdr.messageto]['Hop_Count']
                    self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.BRRP,
                                                                                     message_source,
                                                                                     hdr.messageto,
                                                                                     message_source,
                                                                                     0, hdr.sequencenumber, hdr.rreq_id,
                                                                                     route,
                                                                                     self.componentinstancenumber,
                                                                                     py.messagepayload, py.toNode)))

                else:
                    #print(self.componentinstancenumber, ' not have an entry for the hdr.messageto which is: ', hdr.messageto)
                    pass


        if hdr.messagetype == AODV_ABRMessageTypes.BRRP:
            #print('Inside brrp, I am node', self.componentinstancenumber)
            self.show_routing_table()
            time.sleep(5)

            if hdr.messageto not in self.BBRP_Responses.keys(): #if the bbrq_response[hdr.messageto] is empty, create an entry
                self.BBRP_Responses[hdr.messageto] = {'Int_Sender': str(hdr.int_sender),
                                                      'Hop_Count': str(hdr.hop_count)}
                #self.show_bbrp_responses()

                hopCount = int(self.BBRP_Responses[hdr.messageto]['Hop_Count']) + 1
                hopCount2 = int(self.RoutingTable[py.toNode]['Hop_Count']) + 1

                self.RoutingTable[py.toNode] = {'Dest': str(py.toNode),
                                                'Next_Hop': str(self.BBRP_Responses[hdr.messageto]['Int_Sender']),
                                                'Seq_No': str(hdr.sequencenumber),
                                                'Hop_Count': str(hopCount2)}

                self.RoutingTable[hdr.messageto] = {'Dest': str(hdr.messageto),
                                                'Next_Hop': str(self.BBRP_Responses[hdr.messageto]['Int_Sender']),
                                                 'Seq_No': str(hdr.sequencenumber),
                                                 'Hop_Count': str(hopCount)}
                self.show_routing_table()
            else:
                if int(self.BBRP_Responses[hdr.messageto]['Hop_Count']) > int(hdr.hop_count): # if the received brrp response is better
                    self.BBRP_Responses[py.toNode] = {'Int_Sender': str(hdr.int_sender),
                                                      'Hop_Count': str(hdr.hop_count)}
                    #self.show_bbrp_responses()
                else:
                    #print('if the received brrp response is the same')
                    pass

                hopCount = int(self.BBRP_Responses[hdr.messageto]['Hop_Count']) + 1

                self.RoutingTable[hdr.messageto] = {'Dest': str(hdr.messageto),
                                                'Next_Hop': str(self.BBRP_Responses[hdr.messageto]['Int_Sender']),
                                                 'Seq_No': str(hdr.sequencenumber),
                                                 'Hop_Count': str(hopCount)}

                self.show_routing_table()

        if hdr.messagetype == AODV_ABRMessageTypes.RERR:

            if hdr.messagefrom == self.componentinstancenumber:
                print('I am node ', self.componentinstancenumber, ' and restarting RREQ after receiving RERR'),
                for i in self.NeighbourList:
                    self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.RREQ,
                                                                                     py.messagepayload,
                                                                                     hdr.messageto,
                                                                                     int(i),
                                                                                     0, hdr.sequencenumber+1, hdr.rreq_id +1,
                                                                                     0,
                                                                                     self.componentinstancenumber)))
            else:
                print('hello from RERR, I am node', self.componentinstancenumber, ' and forwarding RERR message to the node ', py.messagepayload)
                self.show_routing_table()
                self.send_down(Event(self, EventTypes.MFRT, self.prepare_message(AODV_ABRMessageTypes.RERR,
                                                                                 py.messagepayload,
                                                                                 hdr.messageto,
                                                                                 int(self.RoutingTable[py.messagepayload]['Next_Hop']),
                                                                                 0, hdr.sequencenumber,
                                                                                 hdr.rreq_id,
                                                                                 hdr.hop_count,
                                                                                 self.componentinstancenumber)))

    def start_routing(self, src_node, dest_node):
        print('Starting AODV-ABR Routing from node', src_node.componentinstancenumber, 'to node ',
              dest_node.componentinstancenumber)
        self.send_self(Event(self, EventTypes.MFRB,
                             self.prepare_message(AODV_ABRMessageTypes.RREQ, self.componentinstancenumber,
                                                  dest_node.componentinstancenumber, self.componentinstancenumber, 0, 1, 1, 1,  0, self.componentinstancenumber)))

    def prepare_message(self, message_type: AODV_ABRMessageTypes, source: int, dest: int, next_hop:int, int_id:int, seq_number:int, rreq_id:int, hop_count:int,
                        int_sender: int, payload: str = None, payload2: str = None) -> GenericMessage:
        header = AODV_ABRMessageHeader(message_type, source, dest, next_hop, int_id, seq_number, rreq_id, hop_count, int_sender )
        payload = AODV_ABRMessagePayload(payload, payload2)
        return GenericMessage(header, payload)

    def show_routing_table(self):
        print('Routing Table of:', self.componentinstancenumber)
        for key, value in self.RoutingTable.items():
            print(key, ' : ', value)

    def show_alternate_route_table(self):
        print('Alternate Route Table of:', self.componentinstancenumber)
        for key, value in self.AlternateRouteTable.items():
            print(key, ' : ', value)

    def show_bbrp_responses(self):
        print('BBRP Responses of: ', self.componentinstancenumber)
        for key, value in self.BBRP_Responses.items():
            print(key, ' : ', value)

    def show_neighbours(self):
        print('Neighbours:')
        for x in range(len(self.NeighbourList)):
            print(self.NeighbourList[x])

    def updateRoutingTableNeighbours(self):
        for x in self.NeighbourList:
            self.RoutingTable[x] = {'Dest': str(x),
                                    'Next_Hop': str(x),
                                    'Seq_No': 100,
                                    'Hop_Count': 1}

    def linkBreak(self, fromNode, toNode, src_node, dest_node):
        print('Link Break Scenario, Link Break between nodes ', fromNode.componentinstancenumber, ' - ', toNode.componentinstancenumber)
        self.send_self(Event(self, EventTypes.MFRB,
                             self.prepare_message(AODV_ABRMessageTypes.BRRQ, fromNode.componentinstancenumber,
                                                  dest_node.componentinstancenumber, fromNode.componentinstancenumber, 0,
                                                  3, 2, 0, fromNode.componentinstancenumber, src_node.componentinstancenumber, toNode.componentinstancenumber)))





import os
import sys
import time
from enum import Enum
sys.path.insert(0, os.getcwd())

from ahc.Ahc import ComponentModel, Event, Topology
from ahc.Ahc import ComponentRegistry
from ahc.Ahc import GenericMessagePayload, GenericMessageHeader, GenericMessage, EventTypes

#TODO: message_count has to be a member variable
message_count = 0
registry = ComponentRegistry()

topo = Topology()

# define your own message types
class ApplicationLayerMessageTypes(Enum):
  PROPOSE = "PROPOSE"
  ACCEPT = "ACCEPT"
  ACCEPT2 = "ACCEPT2"
  CONNECT = "CONNECT"
  INITIATE = "INITIATE"
  TEST = "TEST"
  REPORT = "REPORT"
  REJECT = "REJECT"
  
  
class EdgeStatus(Enum): 
  FIND = "FIND"
  FOUND = "FOUND"

class EdgeType(Enum): 
  BASIC = "BASIC_EDGE"
  BRANCH = "BRANCH_EDGE"
  REJECTED = "REJECTED"

# define your own message header structure
class ApplicationLayerMessageHeader(GenericMessageHeader):
  pass

# define your own message payload structure
class ApplicationLayerMessagePayload(GenericMessagePayload):
  pass

class InitiateMessagePayload:
  def __init__(self, weight, level, status):
    self.weight = weight
    self.level = level
    self.status = status

class TestMessagePayload:
  def __init__(self, fn, level):
    self.fn = fn
    self.level = level

class ReportMessagePayload: 
  def _init_(self, weight): 
    self.weight = weight 

class ConnectMessagePayload:
  def __init__(self, level):
    self.level = level
# class ConnectMessagePayload: 
#   def _init_(self, fn): 
#     self.fn = fn

class ElectionSpiraComponent(ComponentModel):
  def on_init(self, eventobj: Event):
    # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
    pass

  def on_propose(self, eventobj: Event):
    pass
    
    # destination = random.randint(len(Topology.G.nodes))
    # destination = 1
    # hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.PROPOSE, self.componentinstancenumber, destination)
    # payload = ApplicationLayerMessagePayload("23")
    # proposalmessage = GenericMessage(hdr, payload)
    # randdelay = random.randint(0, 5)
    # time.sleep(randdelay)
    # self.send_self(Event(self, "propose", proposalmessage))

  def on_message_from_bottom(self, eventobj: Event):
    # try:
    global message_count
    applmessage = eventobj.eventcontent
    hdr = applmessage.header
    message_count += 1
    if hdr.messagetype == ApplicationLayerMessageTypes.ACCEPT:
      print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
    elif hdr.messagetype == ApplicationLayerMessageTypes.ACCEPT2:
      print(f"Node-{self.componentinstancenumber} is ACCEPTed Node-{hdr.messagefrom}")
      self.accept_message_handler(applmessage.payload, hdr)
    elif hdr.messagetype == ApplicationLayerMessageTypes.CONNECT:
      print(f'Node-{self.componentinstancenumber} wants to be connected with {hdr.messagefrom}')
      self.connect_message_handler(applmessage.payload, hdr)
    elif hdr.messagetype == ApplicationLayerMessageTypes.INITIATE:
      print(f'Node-{self.componentinstancenumber} take a initiate message from {hdr.messagefrom} with values (INITIATE, {applmessage.payload.weight},{applmessage.payload.level},{applmessage.payload.status}') 
      self.initiate_message_handler(applmessage.payload, hdr)
    elif hdr.messagetype == ApplicationLayerMessageTypes.TEST:
      print(f"Node-{self.componentinstancenumber} is TESTed by Node-{hdr.messagefrom}")
      self.test_message_handler(applmessage.payload, hdr)
    elif hdr.messagetype == ApplicationLayerMessageTypes.REJECT:
      print(f"Node-{self.componentinstancenumber} is REJECTed by Node-{hdr.messagefrom}")
      self.reject_message_handler(applmessage.payload, hdr)
    elif hdr.messagetype == ApplicationLayerMessageTypes.REPORT:
      print(f"Node-{self.componentinstancenumber} is REPORTed by Node-{hdr.messagefrom} with weight ===> {applmessage.payload.weight}")
      self.report_message_handler(applmessage.payload, hdr)

    # except AttributeError:
    #   print("Attribute Error")

  def get_edge_weight_with_node(self, node): 
    for i in self.basic_edges + self.branch_edges:
      if i[0] == node or i[1] == node:
        return i[2]
    
    return -1 
  
  def get_edge_with_node(self, node): 
    for i in self.weights:
      if i[0] == node or i[1] == node:
        return i
    return

  def find_lowest_weight_node(self):
    edge = self.basic_edges[0]
    for i in range(1,len(self.basic_edges)):
      if self.basic_edges[i][2] < edge[2]:
        edge = self.basic_edges[i]
      
    return edge 

  def connect_message_handler(self, payload, hdr):
    if payload.level == self.level:
      self.level += 1 
      weight = self.get_edge_weight_with_node(hdr.messagefrom)
      new_hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.INITIATE, self.componentinstancenumber, hdr.messagefrom)
      new_payload = InitiateMessagePayload(weight, self.level, self.status)
      msg = GenericMessage(new_hdr, new_payload)
      self.send_down(Event(self, EventTypes.MFRT, msg))
    elif payload.level < self.level: 
      weight = self.get_edge_weight_with_node(hdr.messagefrom)
      new_hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.INITIATE, self.componentinstancenumber, hdr.messagefrom)
      new_payload = InitiateMessagePayload(weight, self.level, self.status)
      msg = GenericMessage(new_hdr, new_payload)
      self.send_down(Event(self, EventTypes.MFRT, msg))

  def initiate_message_handler(self, payload, hdr): 
    self.fn = payload.weight
    self.status = payload.status 
    self.level = payload.level 
    self.parent = hdr.messagefrom
    # for i in self.branch_edges:
    #   if i[0] == hdr.messagefrom or i[1] == hdr.messagefrom:
    #     pass 
    #   else: 
    #     pass
        # new_hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.INITIATE)
        # new_payload = InitiateMessagePayload(self.fn, self.level, self.status)
        # msg = GenericMessage(new_hdr, new_payload)
        # self.send_down(Event(self, EventTypes.MFRT, msg))

    if payload.status == EdgeStatus.FIND:
      if len(self.basic_edges): 
        test_edge = self.find_lowest_weight_node()
        destination = -1 
        if self.componentinstancenumber == test_edge[0]:
          destination = test_edge[1]
        else: 
          destination = test_edge[0]
        new_hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.TEST, self.componentinstancenumber, destination)
        new_payload = TestMessagePayload(self.fn, self.level)
        msg = GenericMessage(new_hdr, new_payload)
        self.send_down(Event(self, EventTypes.MFRT, msg))
      else: 
        new_hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.REPORT, self.componentinstancenumber, hdr.messagefrom)
        new_payload = ReportMessagePayload(1000)
        msg = GenericMessage(new_hdr, new_payload)
        self.send_down(self, EventTypes.MFRT, msg)
    else: 
      pass 
    
  def test_message_handler(self, payload, hdr): 
    while self.level < payload.level:
      time.sleep(1)
    
    if payload.fn == self.fn:
      new_hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.REJECT, self.componentinstancenumber, hdr.messagefrom)
      new_payload = ApplicationLayerMessagePayload('23')
      msg = GenericMessage(hdr, new_payload)
      self.send_down(Event(self, EventTypes.MFRT, msg))
      return 
    else: 
      new_hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.ACCEPT, self.componentinstancenumber, hdr.messagefrom)
      new_payload = ApplicationLayerMessagePayload('23')
      msg = GenericMessage(hdr, new_payload)
      self.send_down(Event(self, EventTypes.MFRT, msg))

  def reject_message_handler(self, payload, hdr): 
    edge = self.get_edge_with_node(hdr.messagefrom)
    self.rejected_edges.append(edge)
    self.branch_edges.remove(edge)

  def accept_message_handler(self, payload, hdr): 
    weight = self.get_edge_with_node(hdr.messagefrom)
    self.reported_val = weight
    destination = self.parent
    hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.REPORT, self.componentinstancenumber, hdr.messagefrom)
    payload = ReportMessagePayload(weight)
    msg = GenericMessage(hdr, payload)
    self.send_down(Event(self, EventTypes.MFRT, msg))

  def report_message_handler(self, payload, hdr): 
    if self.reported_val > payload.weight:
      pass 
    else: 
      edge = self.find_lowest_weight_node()
      destination = -1
      if edge[1] == self.componentinstancenumber: 
        destination = edge[0]
      else:
        destination = edge[1]
      hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.CONNECT, self.componentinstancenumber, destination)
      payload = ConnectMessagePayload(self.fn)
      msg = GenericMessage(hdr, payload)
      self.send_down(self, EventTypes.MFRT, msg)
    pass 

  def initialize_connect(self): 
    print(f"Node-{self.componentinstancenumber} starts initializing")
    destination = -1
    edge = self.branch_edges[0] 
    if edge[1] == self.componentinstancenumber: 
      destination = edge[0] 
    else: 
      destination = edge[1] 
    hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.CONNECT, self.componentinstancenumber, destination)
    payload = ConnectMessagePayload(self.level)
    msg = GenericMessage(hdr, payload)
    self.send_down(Event(self, EventTypes.MFRT, msg))

  def on_agree(self, eventobj: Event):
    print(f"Agreed on {eventobj.eventcontent}")

  def on_timer_expired(self, eventobj: Event):
    pass


  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    self.eventhandlers["propose"] = self.on_propose
    self.eventhandlers["agree"] = self.on_agree
    self.eventhandlers["timerexpired"] = self.on_timer_expired
    self.weights = []
    self.status = EdgeStatus.FIND

    for x in list(topo.G.edges.data("weight")): 
        if x[1] == componentinstancenumber or x[0] == componentinstancenumber:
          self.weights.append(x)

    self.branch_edges = [] 
    self.rejected_edges = []
    self.basic_edges = self.weights

    #SEND THE LOWEST WEIGHT EDGE TO BRANCH INITIALLY 
    b1_edge = self.find_lowest_weight_node()
    self.branch_edges.append(b1_edge)
    self.basic_edges.remove(b1_edge)

    self.level = 0 
    self.fn = 0 
    self.parent = componentinstancenumber
    self.reported_val = 1000



from ahc.Ahc import Event, EventTypes, ComponentConfigurationParameters
from ahc.MAC.GenericMAC import GenericMac, GenericMacEventTypes
import time, random, math

class MacCsmaPPersistentConfigurationParameters (ComponentConfigurationParameters):
    def __init__(self, p):
        self.p = p

class MacCsmaPPersistent(GenericMac):
    
    #Constructor
    def __init__(self, componentname, componentinstancenumber, configurationparameters:MacCsmaPPersistentConfigurationParameters):
        super().__init__(componentname, componentinstancenumber)
        self.p = configurationparameters.p
    
    #on_init will be called from topo.start to initialize components
    def on_init(self, eventobj: Event):
        self.retrialcnt = 0
        super().on_init(eventobj)  # required because of inheritence
        #print("Initialized", self.componentname, ":", self.componentinstancenumber)

    def handle_frame(self):
        #TODO: not a good solution put message in queue, schedule a future event to retry yhe first item in queueu    
        #print("handle_frame")
        if self.framequeue.qsize() > 0:
            #print("handle_frame", "queue not empty")
            randval = random.random()
            if randval < self.p: # TODO: Check if correct
                clearmi, powerdb  = self.ahcuhd.ischannelclear(threshold=-35)
                #print("Component:", self.componentinstancenumber, "clear mi=", clearmi, " Power=", powerdb)
                if  clearmi == True:
                    try:
                        eventobj = self.framequeue.get()
                        evt = Event(self, EventTypes.MFRT, eventobj.eventcontent)
                        self.send_down(evt)
                        self.retrialcnt = 0
                    except Exception as e:
                        print("MacCsmaPPersistent handle_frame exception, ", e)
                else:
                    self.retrialcnt = self.retrialcnt + 1
                    time.sleep(random.randrange(0,math.pow(2,self.retrialcnt))*0.001)
                    #print("Busy")
        else:
            #print("Queue size", self.framequeue.qsize())
            pass
        time.sleep(0.00001) # TODO: Think about this otherwise we will only do cca
        self.send_self(Event(self, GenericMacEventTypes.HANDLEMACFRAME, None)) #Continuously trigger handle_frame
        
            
                
                


import random
import networkx as nx
import matplotlib.pyplot as plt

class ERG:
    def __init__(self, node_count: int, connectivity: float, ax = None) -> None:
        self.ax = ax
        self.node_count = node_count
        self.root = random.choice(list(range(self.node_count)))
        self.G = nx.nx.erdos_renyi_graph(self.node_count, connectivity, seed=random.randint(100, 10000), directed=False)

    def plot(self):
        node_colors = ["red" if i == self.root else "mediumslateblue" for i in range(self.node_count)]

        if self.ax is not None:
            nx.draw(self.G, with_labels=True, node_color=node_colors, ax=self.ax)
        else:
            nx.draw(self.G, with_labels=True, node_color=node_colors)

class Grid:
    def __init__(self, node_count_on_edge: int, ax = None) -> None:
        self.ax = ax
        self.node_count_on_edge = node_count_on_edge
        self.root = random.choice(list(range(self.node_count_on_edge ** 2)))
        self.G = nx.grid_2d_graph(self.node_count_on_edge, self.node_count_on_edge)
        self.positions = {self.node_count_on_edge * x[0] + x[1]: x for x in self.G.nodes()}
        
        self.G = nx.relabel_nodes(self.G, lambda x: self.node_count_on_edge * x[0] + x[1])
        
    def plot(self):
        node_colors = ["red" if i == self.root else "mediumslateblue" for i in range(self.node_count_on_edge ** 2)]

        if self.ax is not None:
            nx.draw(self.G, with_labels=True, node_color=node_colors, pos=self.positions, ax=self.ax)
        else:
            nx.draw(self.G, with_labels=True, node_color=node_colors, pos=self.positions)

class Star:
    def __init__(self, slave_count: int, master_is_root: bool = True, ax = None) -> None:
        self.ax = ax
        self.slave_count = slave_count
        self.root = 0 if master_is_root else random.choice(list(range(1, slave_count + 1)))

        self.G = nx.Graph()

        self.G.add_node(0)

        for i in range(1, self.slave_count + 1):
            self.G.add_node(i)
            self.G.add_edge(0, i)
        
    def plot(self):
        node_colors = ["red" if i == self.root else "mediumslateblue" for i in range(self.slave_count + 1)]

        if self.ax is not None:
            nx.draw(self.G, with_labels=True, node_color=node_colors, ax=self.ax)
        else:
            nx.draw(self.G, with_labels=True, node_color=node_colors)

if __name__ == "__main__":
    fig, axes = plt.subplots(1, 4)
    fig.set_figheight(5)
    fig.set_figwidth(25)
    fig.tight_layout()

    ERG(10, 0.45, ax=axes[0]).plot()
    Grid(4, ax=axes[1]).plot()
    Star(6, master_is_root=True, ax=axes[2]).plot()
    Star(6, master_is_root=False, ax=axes[3]).plot()

    plt.show()


import time


class GSQQueueElement:
    link_states = {}
    sequence_numbers = {}
    source_id = -1
    transfer_duration = -1

    def __init__(self, source_id, payload):
        self.link_states = payload["link_states"].copy()
        self.sequence_numbers = payload["sequence_numbers"].copy()
        self.source_id = int(source_id)
        self.transfer_duration = time.time() * 1000 - payload["timestamp"]



from ahc.Routing.SSBR.HelperFunctions import messageParser
from ahc.Ahc import ComponentModel, Event, EventTypes, ComponentRegistry
from ahc.Routing.SSBR.HelperFunctions import SSBRRouteReplyMessage, SSBRRouteCompletedMessage

class DRP(ComponentModel):
    def __init__(self, componentname, componentid):
        super(DRP, self).__init__(componentname, componentid)
        self.componentid = componentid
        self.signalStabilityTable = {}
        self.routingTableFlag = False

    def on_init(self, eventobj: Event):
        print(f"{self.componentname} - #{self.componentid} is up.\n")
        #evt = Event(self, EventTypes.MFRP, "Network interface to peers")
        #self.send_peer(evt)

    def on_message_from_bottom(self, eventobj: Event):
        messageto = eventobj.eventcontent.header.messageto
        interfaceID = eventobj.eventcontent.header.interfaceid
        messageFromID = int(interfaceID.split("-")[1])
        #selfFP = ComponentRegistry().get_component_by_key("FP",self.componentid)
        #selfFP.routingTable.get()[int(messageto.split("-")[1])] != None
        if messageFromID == int(self.componentid):
            messageFromID = int(interfaceID.split("-")[0])

        if(eventobj.eventcontent.header.messagetype == "ROUTESEARCH"):
            if self.routingTableFlag == False:
                tableVal = self.signalStabilityTable.get(messageFromID)
                if tableVal == "SC":
                    self.routingTableFlag = True
                    if int(messageto.split("-")[1]) == int(self.componentid):
                        evt = Event(self, EventTypes.MFRP, SSBRRouteReplyMessage(self, eventobj))
                        self.send_peer(evt)
                    else:
                        evt = Event(self, EventTypes.MFRP,messageParser(self, eventobj))
                        self.send_peer(evt)
            

        elif(eventobj.eventcontent.header.messagetype == "ROUTEREPLY"):
                evt = Event(self, EventTypes.MFRP,messageParser(self, eventobj))
                self.send_peer(evt)
            
        else:
            evt = Event(self, EventTypes.MFRP,messageParser(self, eventobj))
            self.send_peer(evt)
    
    def editSignalStabilityTable(self, item, mode):
        if mode == "SC":
            self.signalStabilityTable[item] = "SC"
        else:
            self.signalStabilityTable[item] = "WC"
        
    def printSignalStabilityTable(self):
        print(self.signalStabilityTable)

    def getSignalStabilityTable(self):
        return self.signalStabilityTable



from numpy import inner
from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, GenericMessagePayload, EventTypes, ComponentRegistry, Lock, Thread, Topology, ConnectorTypes, MessageDestinationIdentifiers
from ahc.Broadcasting.Broadcasting import ControlledFlooding, BroadcastingEventTypes, BroadcastingMessageHeader, BroadcastingMessageTypes
from ahc.Channels.Channels import P2PFIFOFairLossChannel, P2PFIFOPerfectChannel, Channel
from ahc.LinkLayers.GenericLinkLayer import LinkLayer

from enum import Enum

from threading import Lock

import random
import time


inf = float('inf')

class AODVNode(ComponentModel):
   
    def on_message_from_top(self, eventobj: Event):
        self.send_down(Event(self, EventTypes.MFRT, eventobj.eventcontent))

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))

    def __init__(self, componentname, componentid):
        # SUBCOMPONENTS
        self.applicationlayer = ApplicationLayer("ApplicationLayer", componentid)
        self.aodvservice = AODVLayer("AODVLayer", componentid)
        self.linklayer = LinkLayer("LinkLayer", componentid)

        # CONNECTIONS AMONG SUBCOMPONENTS
        self.applicationlayer.connect_me_to_component(ConnectorTypes.DOWN, self.aodvservice)
        self.aodvservice.connect_me_to_component(ConnectorTypes.UP, self.applicationlayer)

        self.aodvservice.connect_me_to_component(ConnectorTypes.DOWN, self.linklayer)
        self.linklayer.connect_me_to_component(ConnectorTypes.UP, self.aodvservice)

        # Connect the bottom component to the composite component....
        self.linklayer.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.linklayer)

        super().__init__(componentname, componentid)

class AODVLayerMessageType(Enum):
    RREQ = "RREQ"
    RREP = "RREP"
    RERR = "RERR"
    DATA = "DATA"

class AODVLayer(ComponentModel):
  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    self.eventhandlers[AODVLayerMessageType.RREQ] = self.on_rreq
    self.eventhandlers[AODVLayerMessageType.RREP] = self.on_rrep
    self.eventhandlers[AODVLayerMessageType.RERR] = self.on_rerr
    self.eventhandlers[AODVLayerMessageType.DATA] = self.on_data
    
    self.rreqLock = Lock() #Lock to prevent simultanious access to db
    self.rrepLock = Lock() #Lock to prevent simultanious access to db

    self.RoutingTable = {}
    
  def on_init(self, eventobj: Event):
    self.uniquebroadcastidentifier = 1
    self.broadcastdb = []
    self.rrepdb = {}
    # if self.componentinstancenumber == 8:
      # self.update_routing_table(4,6,3,1)

  def update_topology(self):
    Topology().nodecolors[self.componentinstancenumber] = 'r'
    # Topology().plot()

  def on_rreq(self, eventobj: Event):
    # self.update_topology()
    applicationLayerMessage = eventobj.eventcontent
    destination = applicationLayerMessage.header.messageto
    source = applicationLayerMessage.header.messagefrom
    sequencenumber = applicationLayerMessage.header.sequencenumber
    self.uniquebroadcastidentifier = self.uniquebroadcastidentifier + 1
    interfaceid = self.uniquebroadcastidentifier
    currentComponent = self.componentinstancenumber
    nexthop = MessageDestinationIdentifiers.LINKLAYERBROADCAST

    hopCount = self.get_next_hop_count(source, sequencenumber)

    #print(f"Send RREQ Message from {self.componentname}.{self.componentinstancenumber} for destination {destination} over {nexthop}")
    header = AODVBroadcastingMessageHeader(AODVLayerMessageType.RREQ, currentComponent, destination,
                                    nexthop, interfaceid, hopFrom=currentComponent, hopCount=hopCount + 1)
    
    broadcastmessage = AODVBroadcastingMessage(header, applicationLayerMessage)

    self.send_down(Event(self, EventTypes.MFRT, broadcastmessage))

  def on_rrep(self, eventobj: Event):
    # self.update_topology()
    applicationLayerMessage = eventobj.eventcontent
    destination = applicationLayerMessage.header.messageto
    source = applicationLayerMessage.header.messagefrom
    sequencenumber = applicationLayerMessage.header.sequencenumber
    self.uniquebroadcastidentifier = self.uniquebroadcastidentifier + 1
    interfaceid = self.uniquebroadcastidentifier
    currentComponent = self.componentinstancenumber
    nexthop = self.get_next_hop(source, sequencenumber)
    hopCount = self.get_next_hop_count(destination, sequencenumber)

    # print(f"RREP Message Send # Sender: {self.componentname}.{self.componentinstancenumber} for destination {destination} over {nexthop}")
    header = AODVBroadcastingMessageHeader(AODVLayerMessageType.RREP, currentComponent, destination,
                                    nexthop, interfaceid, hopFrom=currentComponent, hopCount=hopCount + 1)
    
    broadcastmessage = AODVBroadcastingMessage(header, applicationLayerMessage)

    self.send_down(Event(self, EventTypes.MFRT, broadcastmessage))

  def on_data(self, eventobj: Event):
    applicationLayerMessage = eventobj.eventcontent
    destination = applicationLayerMessage.header.messageto
    source = applicationLayerMessage.header.messagefrom
    sequencenumber = applicationLayerMessage.header.sequencenumber
    self.uniquebroadcastidentifier = self.uniquebroadcastidentifier + 1
    interfaceid = self.uniquebroadcastidentifier
    currentComponent = self.componentinstancenumber
    nexthop = self.get_next_hop(destination, sequencenumber)

    # print(f"RREP Message Send # Sender: {self.componentname}.{self.componentinstancenumber} for destination {destination} over {nexthop}")
    header = AODVBroadcastingMessageHeader(AODVLayerMessageType.DATA, currentComponent, destination,
                                    nexthop, interfaceid, hopFrom=currentComponent)
    
    broadcastmessage = AODVBroadcastingMessage(header, applicationLayerMessage)

    self.send_down(Event(self, EventTypes.MFRT, broadcastmessage))

  def on_rerr(self, eventobj: Event):
    pass
        
  def on_message_from_top(self, eventobj: Event):
    applicationLayerMessage = eventobj.eventcontent
    applicationLayerMessageHeader = applicationLayerMessage.header
    applicationLayerMessagePayloadDetail = applicationLayerMessage.payload.messagepayload

    source = applicationLayerMessageHeader.messagefrom
    destination = applicationLayerMessageHeader.messageto
    sequencenumber = applicationLayerMessageHeader.sequencenumber

    #Check if destination is in routing table
    nexthop = self.get_next_hop(destination, sequencenumber)

    if(nexthop == inf):
        #If it does not exist in routing table
        #Create RREQ message and broadcast it
        evt = Event(self, AODVLayerMessageType.RREQ, applicationLayerMessage)
        self.send_self(evt)

    else:
        #If it exists in routing table 
        #Send data to destination
        pass

  def on_message_from_bottom(self, eventobj: Event):
    broadcastingMessage = eventobj.eventcontent

    broadcastingMessageHeader = broadcastingMessage.header
    applicationLayerMessage = broadcastingMessage.payload
    applicationLayerMessageHeader = applicationLayerMessage.header

    source = applicationLayerMessageHeader.messagefrom
    destination = broadcastingMessageHeader.messageto
    previousNode = broadcastingMessageHeader.messagefrom
    hopCount = broadcastingMessageHeader.hopCount
    sequencenumber = applicationLayerMessageHeader.sequencenumber

    if broadcastingMessageHeader.messagetype == AODVLayerMessageType.RREQ:
            
      with self.rreqLock:
        if (applicationLayerMessage.uniqueid in self.broadcastdb):
          print(f"RREQ Message Received Already # Receiver: {self.componentname}.{self.componentinstancenumber} from node: {previousNode} with uniqueid {applicationLayerMessage.uniqueid}. Skipping...")

        else:
          self.broadcastdb.append(applicationLayerMessage.uniqueid)
        
          if(source == self.componentinstancenumber):
            print(f"RREQ Message Arrived to Source # Receiver: {self.componentname}.{self.componentinstancenumber} from node: {previousNode} with uniqueid {applicationLayerMessage.uniqueid}. Skipping...")

          else:
            print(f"RREQ Message Received # Receiver: {self.componentname}.{self.componentinstancenumber} from node:{previousNode} for source:{source}")
            self.update_routing_table(source, previousNode, hopCount, sequencenumber)

            self.update_topology()
            if destination == self.componentinstancenumber:
              #RREQ arrived to destination. Send RREP
              time.sleep(random.randint(1, 1))
              print(f"RREP Message Send # Sender: {self.componentname}.{self.componentinstancenumber} for source {source} for destionation {destination}")
              evt = Event(self, AODVLayerMessageType.RREP, applicationLayerMessage)
              self.send_self(evt)
            
            elif self.get_next_hop(destination, sequencenumber) != inf:
              #Destination is known from routing table, send RREP
              time.sleep(random.randint(1, 1))
              print(f"Node Knows Destination # Receiver: {self.componentname}.{self.componentinstancenumber} for source {source} for destionation {destination}")
              print(f"RREP Message Send # Sender: {self.componentname}.{self.componentinstancenumber} for source {source} for destionation {destination}")
              evt = Event(self, AODVLayerMessageType.RREP, applicationLayerMessage)
              self.send_self(evt)
              
            else:
              #Rebroadcast RREQ
              time.sleep(random.randint(1, 1))
              print(f"RREQ Message Resend # Sender: {self.componentname}.{self.componentinstancenumber} for source {source} for destination {destination}")
              evt = Event(self, AODVLayerMessageType.RREQ, applicationLayerMessage)
              self.send_self(evt)
    
    elif broadcastingMessageHeader.messagetype == AODVLayerMessageType.RREP:

      with self.rrepLock:
        if (applicationLayerMessage.uniqueid in self.rrepdb) and (self.rrepdb[applicationLayerMessage.uniqueid]["hopCount"] <= hopCount):
          print(f"RREP Message Received Already # Receiver: {self.componentname}.{self.componentinstancenumber} from node: {previousNode} with uniqueid {applicationLayerMessage.uniqueid}. Skipping...")

        else:
          self.rrepdb[applicationLayerMessage.uniqueid] = {"hopCount": hopCount}
          self.update_routing_table(destination, previousNode, hopCount, sequencenumber)

          if(source == self.componentinstancenumber):
            print(f"RREP Message Arrived to Source # Receiver: {self.componentname}.{self.componentinstancenumber} from {previousNode} with uniqueid {applicationLayerMessage.uniqueid}")
            print(f"DATA Message Send # Sender: {self.componentname}.{self.componentinstancenumber} for source {source} for destionation {destination}")
            evt = Event(self, AODVLayerMessageType.DATA, applicationLayerMessage)
            self.send_self(evt)
        
          else:
            print(f"Received RREP Message # Receiver: {self.componentname}.{self.componentinstancenumber} from {previousNode} for source {source}")
        
            self.update_topology()
        
            if self.get_next_hop(source, sequencenumber) != inf:
              print(f"RREP Message Send # Sender {self.componentname}.{self.componentinstancenumber} for source {source} for destionation {destination}")
              evt = Event(self, AODVLayerMessageType.RREP, applicationLayerMessage)
              self.send_self(evt)
            else:
              #If it is not exist in routing table
              pass

    elif broadcastingMessageHeader.messagetype == AODVLayerMessageType.DATA:      
    
      if(destination == self.componentinstancenumber):
        print(f"DATA Message Arrived to Destination # Receiver: {self.componentname}.{self.componentinstancenumber} from {previousNode} with uniqueid {applicationLayerMessage.uniqueid}")
        self.send_up(Event(self, EventTypes.MFRB, eventobj.eventcontent))
    
      else:
        print(f"Received DATA Message # Receiver: {self.componentname}.{self.componentinstancenumber} from {previousNode} for destination {destination}")
    
        if self.get_next_hop(destination, sequencenumber) != inf:
          print(f"DATA Message Send # Sender: {self.componentname}.{self.componentinstancenumber} for source {source} for destionation {destination}")
          evt = Event(self, AODVLayerMessageType.DATA, applicationLayerMessage)
          self.send_self(evt)
        else:
          #If it is not exist in routing table
          pass

  def _get_next_hop_data(self, destination, destinationSequenceNumber):
    try:
      #Routing Table->toId, nextHop, toSeqNum, lifeTime
      retval = self.RoutingTable[destination]
      #TODO if timeout is occured, delete record
      #if not, update timeout
      if(retval["destinationSequenceNumber"] >= destinationSequenceNumber):
        return retval
      else:
        return inf
    except KeyError:
      return inf
    except IndexError:
      return inf

  def get_next_hop(self, destination, destinationSequenceNumber):
    retval = self._get_next_hop_data(destination, destinationSequenceNumber)
    if retval != inf:
      return retval["nextHop"]

    return inf

  def get_next_hop_count(self, destination, destinationSequenceNumber):
    retval = self._get_next_hop_data(destination, destinationSequenceNumber)
    if retval != inf:
      return retval["hopCount"]

    return 0

  def update_routing_table(self, destination, nextHop, hopCount, destinationSequenceNumber):
    retval = inf
    try:
      #Routing Table->toId, nextHop, toSeqNum, lifeTime
      retval = self.RoutingTable[destination]
      #TODO if timeout is occured, delete record
      #if not, update timeout
    except KeyError:
      pass
    except IndexError:
      pass
    
    if((retval == inf) or (hopCount < retval["hopCount"])):
      print(f"Routing Table Updated # Node: {self.componentname}.{self.componentinstancenumber} Data => destination:{destination} nextHop:{nextHop} hopCount:{hopCount} sequenceNumber:{destinationSequenceNumber}")
      # lifeTime da gerekli olabilir
      self.RoutingTable[destination] = {
        "destination": destination,
        "nextHop": nextHop,
        "hopCount": hopCount,
        "destinationSequenceNumber": destinationSequenceNumber
      }
    
class ApplicationLayerMessageTypes(Enum):
    PROPOSE = "PROPOSE"
    ACCEPT = "ACCEPT"

# class ApplicationLayerMessageHeader(GenericMessageHeader):
#     pass

# class ApplicationLayerMessagePayload(GenericMessagePayload):
#     pass

class ApplicationLayer(ComponentModel):
    def on_init(self, eventobj: Event):
        print(f"Initializing {self.componentname}.{self.componentinstancenumber}")

        # For testing purpose
        #####################
        if self.componentinstancenumber == 0:
            # destination = random.randint(len(Topology.G.nodes))
            destination = 4
            header = AODVApplicationLayerMessageHeader(ApplicationLayerMessageTypes.PROPOSE, self.componentinstancenumber, destination, sequencenumber=1)
            payload = AODVApplicationLayerMessagePayload("Test Message")
            message = AODVApplicationLayerMessage(header, payload)
            randdelay = random.randint(3, 5)
            time.sleep(randdelay)
          
            print(f"Message Send # Sender: {self.componentname}.{self.componentinstancenumber} to destination {destination}")
            self.send_down(Event(self, EventTypes.MFRT, message))
        ########################

    def on_message_from_top(self, eventobj: Event):
        print(f"{EventTypes.MFRT}  {self.componentname}.{self.componentinstancenumber}")

    def on_message_from_bottom(self, eventobj: Event):
        print(f"Message Received # Node: {self.componentname}.{self.componentinstancenumber} Message: {eventobj.eventcontent.payload.payload.messagepayload}")

class AODVChannel(Channel):
  def on_deliver_to_component(self, eventobj: Event):
    message = eventobj.eventcontent
    linklayerMessageHeader = message.header
    broadcastingMessage = message.payload
    broadcastingMessageHeader = broadcastingMessage.header
    applicationLayerMessage = broadcastingMessage.payload

    messageType = broadcastingMessageHeader.messagetype

    if messageType == AODVLayerMessageType.RREQ:
      callername = eventobj.eventsource.componentinstancenumber
      for item in self.connectors:
        callees = self.connectors[item]
        for callee in callees:
          calleename = callee.componentinstancenumber
          
          if calleename == callername:
            pass
          else:
          
            myevent = Event(eventobj.eventsource, EventTypes.MFRB,
                            eventobj.eventcontent, self.componentinstancenumber,
                            eventid=eventobj.eventid)
            callee.trigger_event(myevent)

    elif messageType == AODVLayerMessageType.RREP or messageType == AODVLayerMessageType.DATA:
      nexthop = broadcastingMessageHeader.nexthop
      callername = eventobj.eventsource.componentinstancenumber
      for item in self.connectors:
        callees = self.connectors[item]
        for callee in callees:
          calleename = callee.componentinstancenumber
          
          if calleename == callername:
            pass
          elif calleename == nexthop:
            myevent = Event(eventobj.eventsource, EventTypes.MFRB,
                            eventobj.eventcontent, self.componentinstancenumber,
                            eventid=eventobj.eventid)
            callee.trigger_event(myevent)

class AODVBroadcastingMessageHeader(BroadcastingMessageHeader):
  def __init__(self, messagetype, messagefrom, messageto, nexthop=float('inf'), interfaceid=float('inf'), sequencenumber=-1, hopFrom=None, hopCount=-1):
    super().__init__(messagetype, messagefrom, messageto, nexthop, interfaceid, sequencenumber)
    self.hopFrom = hopFrom
    self.hopCount = hopCount
  
class AODVBroadcastingMessagePayload(GenericMessagePayload):
  pass

class AODVBroadcastingMessage(GenericMessage):
  pass

class AODVApplicationLayerMessageHeader(GenericMessageHeader):
  pass

class AODVApplicationLayerMessagePayload(GenericMessagePayload):
  pass

class AODVApplicationLayerMessage(GenericMessage):
  pass


from ahc.Ahc import ComponentModel, Event, ConnectorTypes
from ahc.Routing.GSR.GSRExperimentApplicationComponent import GSRExperimentApplicationComponent
from ahc.Routing.GSR.GSRCoordinatorComponent import GSRCoordinator
from ahc.Routing.GSR.GSRMiddlewareStubComponent import GSRMiddlewareMessageStub
from ahc.Routing.GSR.RoutingGSRComponent import RoutingGSRComponent


# Encapsulator for the Application Node
class GSRExperimentNode(ComponentModel):
    def __init__(self, componentname, componentid):
        super(GSRExperimentNode, self).__init__(componentname, componentid)
        self.Application = GSRExperimentApplicationComponent(GSRExperimentApplicationComponent.__name__, componentid)
        self.Coordinator = GSRCoordinator(GSRCoordinator.__name__, componentid)
        self.GSRComponent = RoutingGSRComponent(RoutingGSRComponent.__name__, componentid)
        self.MiddlewareStub = GSRMiddlewareMessageStub(GSRMiddlewareMessageStub.__name__, componentid)

        # Application layer only talks with Coordinator,
        # Coordinator coordinates all demands of the application by triggering other components if necessary...
        self.Application.connect_me_to_component(ConnectorTypes.DOWN, self.Coordinator)
        self.Coordinator.connect_me_to_component(ConnectorTypes.UP, self.Application)
        self.Coordinator.connect_me_to_component(ConnectorTypes.PEER, self.GSRComponent)
        self.Coordinator.connect_me_to_component(ConnectorTypes.DOWN, self.MiddlewareStub)


        self.GSRComponent.connect_me_to_component(ConnectorTypes.PEER, self.Coordinator)

        self.GSRComponent.connect_me_to_component(ConnectorTypes.DOWN, self.MiddlewareStub)
        self.MiddlewareStub.connect_me_to_component(ConnectorTypes.UP, self.GSRComponent)
        self.MiddlewareStub.connect_me_to_component(ConnectorTypes.UP, self.Coordinator)
        self.MiddlewareStub.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.MiddlewareStub)

    def on_init(self, eventobj: Event):
        pass

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(eventobj) # send incoming messages to upper components

    def on_message_from_top(self, eventobj: Event):
        self.send_down(eventobj) # send incoming messages from upper components to a channel




import time

from ahc.Ahc import ComponentModel, Event, GenericMessage, MessageDestinationIdentifiers, GenericMessageHeader, EventTypes, \
    ComponentRegistry, Lock, \
    Thread, Topology
from enum import Enum
import datetime
import random


class CGSREventTypes(Enum):
    HELLO = "HELLO_MESSAGE"
    RREQ = 'ROUTE_REQUEST_PACKET'
    RREP = 'RREP'


class CGSRMessageType(Enum):
    HELLO = "HELLO_MESSAGE"
    RREQ = 'ROUTE_REQUEST_PACKET'
    RREP = 'RREP'
    GENERATE_HELLO = "GENERATE_HELLO"
    GENERATE_RREQ = "GENERATE_RREQ"


class CGSRRouter(ComponentModel):

    def __init__(self, componentname, componentid):
        super(CGSRRouter, self).__init__(componentname, componentid)
        # self.eventhandlers[CGSREventTypes.HELLO] = self.onHello
        self.eventhandlers[CGSREventTypes.RREQ] = self.onRREQ
        self.eventhandlers[CGSREventTypes.RREP] = self.onRREP
        self.neighbor_table = []  # hello mesajinin icinde
        # {'neighbor_id','neighbor_role'}
        self.cluster_adjacency_table = []
        # {neighbor_cluster_head_id,gateway_node_id[]}
        self.two_hop_topology_database = []
        self.message_id_table = []
        self.cluster_addres_list = []
        self.prev_RREQs = []
        self.message_id = 0
        self.sleep = random.randint(1, 5)
        self.node_status = 'C_UNDECIDED'
        self.neighbors = Topology().get_neighbors(self.componentinstancenumber)
        # print('__init__')
        self.send_self(Event(self, EventTypes.INIT, None))

    def on_init(self, eventobj: Event):
        super(CGSRRouter, self).on_init(eventobj)

        self.onNeighbors()
        # self.generateHello()

    def on_message_from_peer(self, eventobj: Event):

        payload = eventobj.eventcontent
        if payload == CGSRMessageType.GENERATE_HELLO:
            sender_status = self.node_status
            neighbor_table = self.neighbor_table
            cat_extension = [] if self.node_status == 'C_MEMBER' else None
            if self.node_status == 'C_MEMBER':
                for item in self.cluster_adjacency_table:
                    cat_extension.append(item['neighbor_cluster_head_id'])
            payload = {'sender_status': sender_status, 'neighbor_table': neighbor_table,
                       'cluster_extension': cat_extension}

            for node in self.neighbors:
                header = GenericMessageHeader(CGSRMessageType.HELLO,
                                              CGSRRouter.__name__ + "-" + str(self.componentinstancenumber),
                                              CGSRRouter.__name__ + "-" + str(node),
                                              interfaceid=str(self.componentinstancenumber) + "-" + str(node))
                eventContent = GenericMessage(header, payload)
                self.send_down(Event(self, EventTypes.MFRT, eventContent))
        if payload == CGSRMessageType.GENERATE_RREQ:
            # print('generate rreq')
            node = Topology().get_random_node()
            payload = {'message_id': self.message_id,
                       'target_address': CGSRRouter.__name__ + "-" + str(node.componentinstancenumber),
                       'num1': len(self.cluster_adjacency_table), 'num2': len(self.cluster_addres_list),
                       'pair_list': self.cluster_adjacency_table
                , 'cluster_list': self.cluster_addres_list,'calculated_route':[CGSRRouter.__name__ + "-" + str(node.componentinstancenumber)],
                       'cluster_addresses':[]}
            header=GenericMessageHeader(CGSRMessageType.RREQ,
                                              CGSRRouter.__name__ + "-" + str(self.componentinstancenumber),
                                              CGSRRouter.__name__ + "-" + str(self.neighbors[0]),
                                              interfaceid=str(self.componentinstancenumber) + "-" + str(self.neighbors[0]))
            eventContent = GenericMessage(header, payload)
            self.send_down(Event(self, EventTypes.MFRT, eventContent))

    def onNeighbors(self):
        # print(self.componentinstancenumber)
        # print(self.neighbors)

        if self.neighbors and self.componentinstancenumber <= self.neighbors[0]:
            self.node_status = "C_HEAD"
        else:
            self.node_status = 'C_MEMBER'
        # print(f"node status for {self.unique_name()}:  " + self.node_status)
        # for node in self.neighbors:
        #     if node>self.neighbors[0]: self.neighbor_table.append({'neighbor_id':node,'neighbor_role':'C_MEMBER'})
        #     else: self.neighbor_table.append({'neighbor_id':node,'neighbor_role':'C_HEAD'})
        # for item in self.neighbor_table:
        #     print(item)

    def on_message_from_bottom(self, eventobj: Event):
        # print('router on message from bottom')
        message_destination = eventobj.eventcontent.header.messageto.split("-")[0]

        # print(f"{self.componentinstancenumber} received {message_destination}")
        if message_destination == CGSRRouter.__name__:  # process only the messages targeted to this component...
            message_source_id = eventobj.eventcontent.header.messagefrom.split("-")[1]
            message_type = eventobj.eventcontent.header.messagetype
            content = eventobj.eventcontent.payload
            print(message_type)
            if message_type == CGSRMessageType.HELLO:
                self.onHello(eventobj)
            if message_type==CGSRMessageType.RREQ:
                # print('rreq geldi')
                self.onRREQ(eventobj)


    def update_cat(self, header, payload):
        sender_id = header.messagefrom
        neighbor_table = payload['neighbor_table']
        summarized_cat = payload['cluster_extension']
        if self.node_status == "C_MEMBER":
            for index, item in enumerate(neighbor_table):
                if item['neighbor_role'] == 'C_HEAD' and item['neighbor_id'] != self.neighbors[0]:
                    self.cluster_adjacency_table.append({'neighbor_cluster_head_id': item['neighbor_id'],
                                                         'gateway_node_id': sender_id})

        if self.node_status == "C_HEAD":
            if any(item.get('neighbor_role') == 'C_HEAD' for item in neighbor_table):
                index = next(i for i, item in enumerate(neighbor_table) if
                             item["neighbor_role"] == 'C_HEAD')
                self.cluster_adjacency_table.append({'neighbor_cluster_head_id': neighbor_table[index]['neighbor_id'],
                                                     'gateway_node_id': sender_id})
            if summarized_cat is not None:
                for entry in summarized_cat:
                    if not any(item.get('neighbor_cluster_head_id') == entry for item in self.cluster_adjacency_table):
                        self.cluster_adjacency_table.append(
                            {'neighbor_cluster_head_id': entry, 'gateway_node_id': sender_id})

    def update_two_hop_topology_database(self, payload):
        neighbor_table = payload['neighbor_table']

        for item in neighbor_table:
            if not any(element.get('neighbor_id') == item['neighbor_id'] for element in self.neighbor_table):
                self.two_hop_topology_database.append(item['neighbor_id'])

    def onHello(self, eventobj: Event):
        # time.sleep(1)
        # print(f"neighbor table for {self.unique_name()}:  ")
        # print(self.neighbor_table)

        message = eventobj.eventcontent
        message_header = message.header
        message_payload = message.payload

        self.update_cat(message_header, message_payload)
        self.update_two_hop_topology_database(message_payload)
        if not any(item.get('neighbor_id') == message_header.messagefrom for item in self.neighbor_table):

            self.neighbor_table.append(
                {'neighbor_id': message_header.messagefrom, 'neighbor_role': message_payload['sender_status']})
        else:
            pass

    def onRREQ(self, eventobj: Event):
        message = eventobj.eventcontent
        message_header = message.header
        message_payload = message.payload
        if message_payload['target_address']==CGSRRouter.__name__+'-'+str(self.componentinstancenumber):
            message.payload['calculated_route'].append(CGSRRouter.__name__+'-'+str(self.componentinstancenumber))
            self.send_self(Event(self,CGSREventTypes.RREP,message))
            return
        self.prev_RREQs.append(message)
        if self.node_status == "C_MEMBER":
            if any(neighbor.get('neighbor_id') == message_payload['target_address'] for neighbor in self.neighbor_table):

                message.header.messageto=message_payload['target_address']
                self.send_down(Event(self, EventTypes.MFRT, message))
            else:
                if any(item.get('gateway_node_address') == CGSRRouter.__name__+'-'+str(self.componentinstancenumber) for item in
                       message_payload['pair_list']):
                    message_header.messageto = self.neighbors[0]
                    msg = GenericMessage(message_header, message_payload)
                    self.send_down(Event(self, CGSRMessageType.RREQ, msg))
                else:
                    return
        elif self.node_status == "C_HEAD":
            if message_payload['message_id'] in self.message_id_table:
                return
            else:
                message_payload['cluster_list'].append(CGSRRouter.__name__+'-'+str(self.componentinstancenumber))
            if any(neighbor.get('neighbor_id') == message_payload['target_address'] for neighbor in
                   self.neighbor_table) or message_payload['target_address'] in self.two_hop_topology_database:
                message.header.messageto = message_payload['target_address']
                self.send_down(Event(self, EventTypes.MFRT, message))
                # uni-cast RREQ to D

            else:
                arr = []
                for item in self.cluster_adjacency_table:
                    arr.append(item['neighbor_cluster_head_id'])
                for index, CH in enumerate(arr):
                    if CH in self.prev_RREQs:
                        continue
                    elif CH in message_payload['cluster_list']:
                        continue
                    else:
                        message_payload['pair_list'].append(
                            {self.prev_RREQs[index]['gateway_node'], self.prev_RREQs[index]['adj_cls_id']})


    def onRREP(self, eventobj: Event):
        message = eventobj.eventcontent
        message_header = message.header
        message_payload = message.payload
        sent_message_payload={}
        sent_message_header={}
        print('onRREP')
        sent_to_x = False
        if self.node_status == 'C_HEAD':
            sent_message_payload['num1']= message_payload['num1'] - 1
            # cluster address[num1]in gateway nodeunu bulacagiz, mesaji node'a forwardla
            cluster_head_id = message_payload[message_payload['num1']]
            calculated_route = message_payload['calculated_route'][message_payload['num2']]
            if any(item.get('neighbor_cluster_head_id') == cluster_head_id for item in self.cluster_adjacency_table):
                index = next(i for i, item in enumerate(self.cluster_adjacency_table) if
                             item["neighbor_cluster_head_id"] == cluster_head_id)
                for gateway_node in self.cluster_adjacency_table[index]['gateway_node_id']:
                    if any(item.get('2hop') == calculated_route and item.get('neighbor') == gateway_node for
                           item in self.two_hop_topology_database):
                        sent_to_x = True
                        message_header.messageto = "CGSRouter"+'-'+gateway_node
                        msg = GenericMessage(message_header, message_payload)
                        self.send_down(Event(self, CGSRMessageType.RREP, msg))

                if not sent_to_x:
                    message_payload['num2'] += 1
                    message_payload['calculated_route'].append(CGSRRouter.__name__+'-'+self.componentinstancenumber)
        elif self.node_status == "C_MEMBER":
            message_payload['num2'] += 1
            message_payload['calculated_route'].append(CGSRRouter.__name__+'-'+self.componentinstancenumber)
            if any(item.get('neighbor_id') == message_payload['cluster_address_list'][message_payload['num1']]
                   for item in self.neighbor_table):
                message_header.messageto = message_payload['cluster_address_list'][
                    message_payload['num1']]
                msg = GenericMessage(message_header, message_payload)
                self.send_down(Event(self, CGSRMessageType.RREP, msg))
            elif any(item.get('neighbor_cluster_head_id') == message_payload['cluster_address_list'][
                message_payload['num1']]
                     for item in self.cluster_adjacency_table):
                index = next(i for i, item in enumerate(self.cluster_adjacency_table) if
                             item["neighbor_cluster_head_id"] == message_payload['cluster_address_list'][
                                 message_payload['num1']])
                message_header.messageto = message_payload['cluster_adjacency_table'][index]
                msg = GenericMessage(message_header, message_payload)
                self.send_down(Event(self, CGSRMessageType.RREP, msg))
                # send rrep to self.cluster_adjacency_table[index]['gateway_node_id']den birine


class HelloMessage(GenericMessage):
    # neighbor_table={'neighbor_id','neighbor_role'}
    # cluster_extension=['adj cluster head address']
    # payload={'sender_status':'member/head','neighbor_table':[],'ClusterAdjacencyExtension':[]}
    def __init__(self, header, payload):
        super(HelloMessage, self).__init__(header, payload)


class ClusterAdjacencyExtension(GenericMessage):
    """
    sadece member nodelar ekler
    helloya extent
    adj. cluster head1 Ip add.
    ....
    """

    def __init__(self, payload):
        super(ClusterAdjacencyExtension, self).__init__(GenericMessageHeader(), payload)


class RouteRequestPacket(GenericMessage):
    """
    ilk RREQ mesaji broadcast
    sadece headlere bu mesaj flood, route discovery
    target address
    message id: unique msg id
    pairlar: arr[num1]: (gateway node, adjacent cluster address(id)-> kendi cluster head id ve CAT'ten diger idleri aliyor)
             arr[num2]: cluster addresses
    headerda destination node adresi: messageto
    """

    payload = {}

    def __init__(self, payload):
        super(RouteRequestPacket, self).__init__(GenericMessageHeader(), payload)


class RREPPacket(GenericMessage):
    """
    msg id, RREQdan gelen idyi buraya kopyala d node yapar
    cluster adres list[num1] ???? sayfa 16
    calculated rute[num2] ???? ch hesapliyor xd
    """

    def __init__(self, payload):
        super(RREPPacket, self).__init__(GenericMessageHeader(), payload)


class SourceRouting(GenericMessage):
    # num,
    # current num,currently visited address,
    # address[num]
    def __init__(self, payload):
        super(SourceRouting, self).__init__(GenericMessageHeader(), payload)



from timeit import default_timer as timer
import random
import time

from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes, Lock, Thread
from ahc.Routing.GSR.Constants import GSR_APPLICATION_NAME, \
    GSR_COORDINATOR_NAME, \
    ROUTING_COMPLETED_MESSAGE_TYPE, \
    TERMINATE_MESSAGE_TYPE, \
    INFO_MESSAGE_TYPE, \
    TOMBALA_DRAW_PERIOD_SECS, \
    TOMBALA_PLAYER_PERIOD_SECS, \
    TOMBALA_MAX_NUMBER, \
    TOMBALA_N_NUMBERS_IN_HAND
from ahc.Routing.GSR.GSRExperimentDataCollector import GSRExperimentCollector

class GSRExperimentApplicationComponent(ComponentModel):
    tombala_message_kind = "TOMBALA"
    winner_message_kind = "WINNER"
    new_number_message_kind = "NEW_NUMBER"

    def __init__(self, componentname, componentid):
        super(GSRExperimentApplicationComponent, self).__init__(componentname, componentid)
        self.n_nodes = 0
        self.game_completed = False
        self.winner = 0
        self.numbers = []
        self.message_queue = []
        self.queue_lock = Lock()

    def on_init(self, eventobj: Event):
        if self.componentinstancenumber > 0:
            self.numbers = [i for i in range(1, TOMBALA_MAX_NUMBER)]
            random.shuffle(self.numbers)
            self.numbers = self.numbers[0:TOMBALA_N_NUMBERS_IN_HAND]
            print(f"Hand of Player {self.componentinstancenumber}: {self.numbers}")
        thread = Thread(target=self.job, args=[14, 4, 18])
        thread.start()

    def on_message_from_bottom(self, eventobj: Event):
        event_header = eventobj.eventcontent.header
        message_to = event_header.messageto.split("-")[0]
        message_type = event_header.messagetype
        message = eventobj.eventcontent.payload
        if message_to == GSR_APPLICATION_NAME:
            if message_type == INFO_MESSAGE_TYPE:
                content = message["content"]
                self.queue_lock.acquire()
                self.message_queue.append(content)
                self.queue_lock.release()
            elif message_type == ROUTING_COMPLETED_MESSAGE_TYPE:
                self.end_time = timer()
                if self.n_nodes == 0:
                    print(f"App {self.componentinstancenumber} has received RoutingCompleted message")
                    self.n_nodes = message["n_nodes"]
                    print(f"There are {self.n_nodes} active nodes in the game")

    def send_info_message(self, dest, payload):
        message_header = GenericMessageHeader(
            INFO_MESSAGE_TYPE,
            GSR_APPLICATION_NAME + "-" + str(self.componentinstancenumber),
            GSR_COORDINATOR_NAME + "-" + str(self.componentinstancenumber))
        payload = {
            "src": self.componentinstancenumber,
            "dest": dest,
            "content": payload
        }
        message = GenericMessage(message_header, payload)
        event = Event(self, EventTypes.MFRT, message)
        self.send_down(event)

    def send_terminate_message(self):
        message_header = GenericMessageHeader(
            TERMINATE_MESSAGE_TYPE,
            GSR_APPLICATION_NAME + "-" + str(self.componentinstancenumber),
            GSR_COORDINATOR_NAME + "-" + str(self.componentinstancenumber))
        message = GenericMessage(message_header, "")
        event = Event(self, EventTypes.MFRT, message)
        self.send_down(event)

    def job(self, *arg):
        if self.componentinstancenumber == 0:
            while not self.game_completed:
                time.sleep(TOMBALA_DRAW_PERIOD_SECS)
                if self.n_nodes == 0:
                    continue
                self.check_messages_master()
                if self.winner > 0:
                    self.announce_winner()
                else:
                    self.announce_new_number()
        else:
            while not self.game_completed:
                time.sleep(TOMBALA_PLAYER_PERIOD_SECS)
                self.check_messages_player()
            if self.winner == self.componentinstancenumber:
                print(f"Player {self.componentinstancenumber}: I'm the winner!")
            else:
                print(f"Player {self.componentinstancenumber} congratulates Player {self.winner}! "
                      f"Numbers left in hand: {self.numbers}")
        self.send_terminate_message()
        GSRExperimentCollector().completion_lock.acquire()
        GSRExperimentCollector().completion.append(self.componentinstancenumber)
        GSRExperimentCollector().completion_lock.release()

    def check_messages_master(self):
        self.queue_lock.acquire()
        for message in self.message_queue:
            if message["kind"] == self.tombala_message_kind:
                winner = message["winner"]
                print(f"Master: Game is over! Player {winner} got TOMBALA!")
                self.game_completed = True
                self.winner = winner
        self.queue_lock.release()

    def check_messages_player(self):
        self.queue_lock.acquire()
        for message in self.message_queue:
            message_kind = message["kind"]
            if message_kind == self.new_number_message_kind:
                new_number = message["number"]
                if new_number in self.numbers:
                    print(f"Player {self.componentinstancenumber}: Has number {new_number}!")
                    self.numbers.remove(new_number)
                    if len(self.numbers) == 0:
                        self.announce_tombala()
            if message_kind == self.winner_message_kind:
                winner = message["winner"]
                self.winner = winner
                self.game_completed = True
        self.queue_lock.release()

    def announce_winner(self):
        for i in range(1, self.n_nodes):
            payload = {
                "kind": self.winner_message_kind,
                "winner": self.winner
            }
            self.send_info_message(i, payload)

    def announce_new_number(self):
        new_number = random.randint(1, TOMBALA_MAX_NUMBER)
        all_player_indexes = [i for i in range(1, self.n_nodes)]
        random.shuffle(all_player_indexes)
        payload = {
            "kind": self.new_number_message_kind,
            "number": new_number,
        }
        print(f"Master: Announcing new number {new_number}!")
        for player_index in all_player_indexes:
            self.send_info_message(player_index, payload)

    def announce_tombala(self):
        print(f"Player {self.componentinstancenumber}: Announcing TOMBALA!")
        payload = {
            "kind": self.tombala_message_kind,
            "winner": self.componentinstancenumber
        }
        self.send_info_message(0, payload)



#!/usr/bin/env python
"""
    Implementation of the AODV Routing Algorithm.
"""

__author__ = "Semih Kurt"
__contact__ = "semikurt@gmail.com"
__copyright__ = "Copyright 2022, WINSLAB"
__credits__ = ["Semih Kurt"]
__date__ = "2022/01/05"
__deprecated__ = False
__email__ = "semikurt@gmail.com"
__license__ = "GPLv3"
__maintainer__ = "developer"
__status__ = "Production"
__version__ = "0.0.1"

import os
import sys
import time
import random

sys.path.insert(0, os.getcwd())

from ahc.Ahc import ComponentModel, Event, ConnectorTypes, Topology
from ahc.Ahc import ComponentRegistry
from ahc.Ahc import GenericMessagePayload, GenericMessageHeader, GenericMessage, EventTypes
from ahc.Channels.Channels import P2PFIFOPerfectChannel
from ahc.LinkLayers.GenericLinkLayer import LinkLayer
from ahc.Routing.AllSeeingEyeNetworkLayer import AllSeingEyeNetworkLayer


class RoutingAODVComponent(ComponentModel):
    def __init__(self,componentname,componentid):
        super().__init__(componentname,componentid)

    def on_init(self, eventobj: Event):
        super().on_init(eventobj)
        
    def on_message_from_bottom(self,eventobj: Event):
        print(f"{self.componentinstancenumber} received")
        



#TODO: Component type and hardwired instance numbers are to be corrected
#TODO: Not a good implementation, to be removed
from enum import Enum
import time
from ahc.Ahc import ComponentModel, Event, GenericMessageHeader, GenericMessagePayload, GenericMessage, EventTypes
from threading import Thread

flag = [False,False]
turn = 0

class PetersonMessageTypes(Enum):
  INC = "+"
  DEC = "-"

class ResourceComponent(ComponentModel):
  def on_init(self, eventobj: Event):
    print(f"Initializing {self.componentname}.{self.componentinstancenumber}\n")
    pass

  def on_message_from_bottom(self, eventobj: Event):
    messagetype = eventobj.eventcontent.header.messagetype ## INC / DEC
    if messagetype == PetersonMessageTypes.INC:
          print("Increment value")
          self.value += 1
    elif messagetype == PetersonMessageTypes.DEC:
          print("Decrement value")
          self.value -= 1
          
  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)
    self.value = 0

  def __repr__(self):
    return f"Value:{self.value}"

class ProducerConsumerComponent(ComponentModel):
  def on_init(self, eventobj: Event):
    print(f"Initializing {self.componentname}.{self.componentinstancenumber}\n")
    pass

  def send_message(self):
    global flag,turn
    message_header = GenericMessageHeader(self.messageType, self.componentinstancenumber, None,interfaceid=f"{self.componentinstancenumber}-{self.resourceComponentId}")
    message_payload = GenericMessagePayload(None)
    message = GenericMessage(message_header, message_payload)
    otherThreadId = 1 - self.id
    for i in range(self.count):
      flag[self.id] = True         # I want to get in to critical section
      turn = otherThreadId         # Lets wait until other thread finish
      while flag[otherThreadId] and turn == otherThreadId:
        time.sleep(0.0000001)
        continue
      ####critical section####
      self.send_down(Event(self, EventTypes.MFRT, message))        
      time.sleep(0.001)
      ########################    
      flag[self.id] = False    # My job is finished

    self.done = True

  def __init__(self, componentname, componentinstancenumber,type = PetersonMessageTypes.INC ,count = 100, resourceComponentId = 2):
    super().__init__(componentname, componentinstancenumber)
    self.count = count
    self.resourceComponentId = resourceComponentId
    self.done = False
    self.messageType = type
    self.id = self.componentinstancenumber

  def start(self):
    t = Thread(target=self.send_message)
    t.daemon = True
    t.start()



from ahc.Routing.SSBR.HelperFunctions import messageParser
from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes

class NetworkInterface(ComponentModel):
    def __init__(self, componentname, componentid):
        super(NetworkInterface, self).__init__(componentname, componentid)
        self.componentid = componentid
        self.RoutingTable = {}
        self.Response_Record = {}

    def on_init(self, eventobj: Event):
        print(f"{self.componentname} - #{self.componentid} is up.\n")
        #evt = Event(self, EventTypes.MFRP, "Network interface to peers")
        #self.send_peer(evt)

    def on_message_from_top(self, eventobj: Event):
        evt = Event(self, EventTypes.MFRT, messageParser(self, eventobj))
        self.send_down(evt)

    def on_message_from_bottom(self, eventobj: Event):
        evt = Event(self, EventTypes.MFRB, messageParser(self, eventobj))
        self.send_up(evt)





from setuptools import setup, find_packages


with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="adhoccomputing",
    version='1.2.0',
    author="Ertan Onur",
    author_email="eronur@metu.edu.tr",
    description="A small example package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cengwins/ahc",
    project_urls={
        "Bug Tracker": "https://github.com/cengwins/ahc/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(),
    python_requires=">=3.8",
)


from ahc.Ahc import ComponentModel, Event, GenericMessage, GenericMessageHeader, EventTypes, ComponentRegistry, Lock, Thread, Topology
import time

class TouegRoutingComponent(ComponentModel):
    def __init__(self, componentname, componentid):
        super(TouegRoutingComponent, self).__init__(componentname, componentid)
        # two dictionaries are indexed with the component id, hence, while broadcasting, the other nodes can easily understant whose distance information they are currently working
        self.DistanceInformation = {self.componentinstancenumber: {}} # stores the shortest path distance values
        self.ParentInformation = {self.componentinstancenumber: {}}
        self.all_process_ids = []
        self.Su = [] # processed node list, algorithm terminates when all nodes are processed...
        self.neighbors = [] # the list of neighbors (ids) connected to main node...
        self.message_queue = [] # for the next invication clear it...
        self.queue_lock = Lock()

    def on_init(self, eventobj: Event):
        super(TouegRoutingComponent, self).on_init(eventobj)
        # the first process does not start immediate, it stars with a peer message
        if self.componentinstancenumber != 0:
            thread = Thread(target=self.job, args=[45, 54, 123])
            thread.start()

    def on_message_from_bottom(self, eventobj: Event):
        message_destination = eventobj.eventcontent.header.messageto.split("-")[0]
        if message_destination == TouegRoutingComponent.__name__: # process only the messages targeted to this component...
            message_source_id = eventobj.eventcontent.header.messagefrom.split("-")[1]
            message_type = eventobj.eventcontent.header.messagetype
            content = eventobj.eventcontent.payload

            if message_type == "INFO" or message_type == "DISTANCE":
                self.queue_lock.acquire() # protect message_queue, both component thread and Toueg thread are trying to access data
                self.message_queue.append((int(message_source_id), message_type, content))
                self.queue_lock.release()


    def on_message_from_peer(self, eventobj: Event):
        message_header = eventobj.eventcontent.header
        message_target = eventobj.eventcontent.header.messageto.split("-")[0]
        if message_target == "TouegRoutingComponent":
            if self.componentinstancenumber == 0:
                if message_header.messagetype == "INITIATEROUTE":
                    thread = Thread(target=self.job, args=[45, 54, 123])
                    thread.start()

    def job(self, *arg):
        self.all_process_ids = []
        for element in ComponentRegistry().components:
            if "MachineLearningNode" in element:
                parts = int(element.split("MachineLearningNode")[1])
                self.all_process_ids.append(parts)
        print("Available nodes : ", self.all_process_ids)
        self.neighbors = Topology().get_neighbors(self.componentinstancenumber) # retrieve all neighbor ids...

        self.neighbor_weights = {a: 1 for a in self.neighbors} # for the time being each edge weight is 1...

        neighbor_ids = [a for a in self.neighbors]
        # found shortest path information will be sent to Coordinator component
        message_payload = self.TOUEG(self.all_process_ids, neighbor_ids, self.neighbor_weights)
        message_header = GenericMessageHeader("ROUTINGCOMPLETED", self.componentname+"-"+str(self.componentinstancenumber),
                                              "Coordinator-"+str(self.componentinstancenumber))
        message = GenericMessage(message_header, message_payload)
        event = Event(self, EventTypes.MFRP, message)
        self.send_peer(event)


    def TOUEG(self, vertices, neigbors, neighbor_weights):
        self.process_id = self.componentinstancenumber
        self.Su = set([])
        self.ParentInformation = {self.process_id: {}}

        for v in vertices:
            if v == self.process_id:
                self.DistanceInformation[self.process_id][v] = 0
                self.ParentInformation[self.process_id][v] = v
            elif v in neigbors:
                self.DistanceInformation[self.process_id][v] = neighbor_weights[v]
                self.ParentInformation[self.process_id][v] = v;
            else:
                self.DistanceInformation[self.process_id][v] = float("inf")
                self.ParentInformation[self.process_id][v] = None

        # For pivot selection, nodes are labeled with their process id
        unordered_vertices = [a for a in vertices]
        unordered_vertices.sort()
        sorted_ids = unordered_vertices
        current_pivot_index = 0
        vertices = set(vertices)
        print(vertices.difference(self.Su))
        while len(vertices.difference(self.Su)) != 0 : # Su != Vertices should be...
            pivot = sorted_ids[current_pivot_index]
            # print(f"Process {self.process_id} picks pivot={pivot}")
            for neighbor in neigbors:
                if self.ParentInformation[self.process_id][pivot] == neighbor:
                    self.sendMessageToNeighbor(neighbor, "INFO", "Child("+str(pivot)+")")
                else:
                    self.sendMessageToNeighbor(neighbor, "INFO", "NotChild("+str(pivot)+")")
            # wait for a specific number of messages
            while True:
                t = self.getPendingChildMessageCount(pivot)
                if t != len(neigbors):
                    time.sleep(0.4)
                else:
                    break

            if self.DistanceInformation[self.process_id][pivot] < float("inf"):
                if self.process_id != pivot:

                    D_pivot = self.waitPivotDistanceFrom(self.ParentInformation[self.process_id][pivot], pivot)
                    while D_pivot is None:
                        D_pivot = self.waitPivotDistanceFrom(self.ParentInformation[self.process_id][pivot], pivot)

                    for neighbor in neigbors:
                        if self.getParticularChildMessage(neighbor, pivot):
                            self.sendMessageToNeighbor(neighbor, "DISTANCE", (pivot, D_pivot))
                    for vertex in vertices:
                        if self.DistanceInformation[self.process_id][vertex] > self.DistanceInformation[self.process_id][pivot] + D_pivot[pivot][vertex]:
                            self.DistanceInformation[self.process_id][vertex] = self.DistanceInformation[self.process_id][pivot]+D_pivot[pivot][vertex]
                            self.ParentInformation[self.process_id][vertex] = self.ParentInformation[self.process_id][pivot]
                elif self.process_id == pivot:
                    received_child_messages = []
                    for neighbor in neigbors:
                        if self.getParticularChildMessage(neighbor, pivot):
                            received_child_messages.append(neighbor)
                    for neighbor in received_child_messages:
                        self.sendMessageToNeighbor(neighbor, "DISTANCE", (pivot, self.DistanceInformation))

            self.Su.add(pivot)
            current_pivot_index += 1
        print(f"\n\nPath Finding has been completed {self.process_id} - {self.DistanceInformation} - {self.ParentInformation}")
        return (self.DistanceInformation, self.ParentInformation)


    def sendMessageToNeighbor(self, neighbor_id, message_type, message):
        message_header = GenericMessageHeader(message_type, TouegRoutingComponent.__name__+"-"+str(self.componentinstancenumber),
        TouegRoutingComponent.__name__+"-"+str(neighbor_id), interfaceid=str(self.componentinstancenumber)+"-"+str(neighbor_id))
        mess_ = GenericMessage(message_header, message)

        event = Event(self, EventTypes.MFRT, mess_)
        self.send_down(event)

    def getPendingChildMessageCount(self, pivot):
        child_message_count = 0
        for i in self.message_queue:
            if i[1] == "INFO" and (("Child(" + str(pivot) + ")" == i[2]) or ("NotChild(" + str(pivot) + ")" == i[2])):
                child_message_count += 1
        return child_message_count

    def waitPivotDistanceFrom(self, source, pivot):
        self.queue_lock.acquire()
        for index, i in enumerate(self.message_queue):
            if i[0] == source and i[1] == "DISTANCE" and i[2][0] == pivot:
                data = self.message_queue.pop(index)
                self.queue_lock.release()
                return data[2][1]
        self.queue_lock.release()
        return None

    def getParticularChildMessage(self, neigh, pivot):
        self.queue_lock.acquire()
        for index, i in enumerate(self.message_queue):
            if i[0] == neigh and i[1] == "INFO" and "Child("+str(pivot)+")" == i[2]:
                data = self.message_queue.pop(index)
                self.queue_lock.release()
                return True
        self.queue_lock.release()
        return False




from enum import Enum
from ahc.Ahc import ComponentModel, Event
from ahc.Ahc import ComponentRegistry
from ahc.Ahc import GenericMessagePayload, GenericMessageHeader, GenericMessage, EventTypes
from ahc.Ahc import Topology
registry = ComponentRegistry()

# define your own message types
class ApplicationLayerMessageTypes(Enum):
  PROPOSE = "PROPOSE"
  ACCEPT = "ACCEPT"
  WAVE = "WAVE"
  ACCEPT_WAVE = "ACCEPT_WAVE"
  FINISH_WAVE = "FINISH_WAVE"

#TODO: message_count can be member variable...
message_count = 0
# define your own message header structure
class ApplicationLayerMessageHeader(GenericMessageHeader):
  pass

# define your own message payload structure
class ApplicationLayerMessagePayload(GenericMessagePayload):
  pass

class WaveMessagePayload:
  def __init__(self, tag):
    self.tag = tag

topo = Topology()

class ElectionEchoExtinctionComponent(ComponentModel):
  def on_init(self, eventobj: Event):
    # print(f"Initializing {self.componentname}.{self.componentinstancenumber}")
    self.neighbors = topo.G.neighbors(self.componentinstancenumber)
    
    # destination = random.randint(len(Topology.G.nodes))
    # destination = 1
    # hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.PROPOSE, self.componentinstancenumber, destination)
    # payload = ApplicationLayerMessagePayload("23")
    # proposalmessage = GenericMessage(hdr, payload)
    # randdelay = random.randint(0, 5)
    # time.sleep(randdelay)
    # self.send_self(Event(self, "propose", proposalmessage))

  def on_message_from_bottom(self, eventobj: Event):
    try:
      applmessage = eventobj.eventcontent
      hdr = applmessage.header
      global message_count 
      message_count += 1
      if hdr.messagetype == ApplicationLayerMessageTypes.ACCEPT:
        print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
      elif hdr.messagetype == ApplicationLayerMessageTypes.PROPOSE:
        print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} has sent {hdr.messagetype} message")
      elif hdr.messagetype == ApplicationLayerMessageTypes.WAVE:
        print(f"Node-{self.componentinstancenumber} get message from Node-{hdr.messagefrom} with tag {applmessage.payload.tag}")
        self.wave_message(applmessage.payload, hdr)
      elif hdr.messagetype == ApplicationLayerMessageTypes.ACCEPT_WAVE:
        print(f"Node-{self.componentinstancenumber} says Node-{hdr.messagefrom} is ACCEPT_WAVE")
        self.accept_wave_message(applmessage.payload, hdr)


    except AttributeError:
      print("Attribute Error")

  # print(f"{self.componentname}.{self.componentinstancenumber}: Gotton message {eventobj.content} ")
  # value = eventobj.content.value
  # value += 1
  # newmsg = MessageContent( value )
  # myevent = Event( self, "agree", newmsg )
  # self.trigger_event(myevent)

  def on_propose(self, eventobj: Event):
    destination = 1
    hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.ACCEPT, self.componentinstancenumber, destination)
    payload = ApplicationLayerMessagePayload("23")
    proposalmessage = GenericMessage(hdr, payload)
    self.send_down(Event(self, EventTypes.MFRT, proposalmessage))

  def on_agree(self, eventobj: Event):
    print(f"Agreed on {eventobj.eventcontent}")

  def on_timer_expired(self, eventobj: Event):
    pass

#TODO: If you call this before all on_inits, then things will go wrong...
  def initiate_process(self):
    self.neighbors = topo.G.neighbors(self.componentinstancenumber)
    print(f"Process initiated {self.componentinstancenumber}")
    self.initiated = True
    for i in self.neighbors:
      destination = i
      hdr = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.WAVE, self.componentinstancenumber, destination)
      payload = WaveMessagePayload(self.componentinstancenumber)
      wave_msg = GenericMessage(hdr, payload)
      self.send_down(Event(self, EventTypes.MFRT, wave_msg))

    
#TO DO WRITE CHECK EVERY  ACCEPT_WAV E MESSAGES CAME OR NOT IN ANOTTHER FUNC

  def accept_wave_message(self, payload, hdr):
    self.waitingAccepts.remove(hdr.messagefrom)
    if len(self.waitingAccepts) == 0:
      destination = self.parent 
      hdr1 = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.ACCEPT_WAVE, self.componentinstancenumber, destination)
      payload = WaveMessagePayload(self.parent)
      wave_msg = GenericMessage(hdr1, payload)
      self.send_down(Event(self, EventTypes.MFRT, wave_msg))

  def wave_message(self, payload, hdr):
    if self.initiated: 
      if payload.tag  > self.parent: 
        self.isWaiting = True
        self.parent = hdr.messagefrom

        for i in self.neighbors:
          destination = i 
          hdr1 = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.WAVE, self.componentinstancenumber, destination)
          wave_msg = GenericMessage(hdr1, payload)
          self.send_down(Event(self, EventTypes.MFRT, wave_msg))
          self.waitingAccepts.append(i)
        
      elif self.parent > payload.tag: 
        print(f"Node-{self.componentinstancenumber} says tag=={payload.tag} from Node-{hdr.messagefrom} is not accepted")
        pass
      else: 
        destination = hdr.messagefrom
        hdr1 = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.ACCEPT_WAVE, self.componentinstancenumber, destination)
        wave_msg = GenericMessage(hdr, payload)
        self.send_down(Event(self, EventTypes.MFRT, wave_msg))
    else: 
      if self.parent != payload.tag:
        self.parent = hdr.messagefrom
        print(f"Node-{self.componentinstancenumber} has new parent Node-{self.parent}")
        for i in self.neighbors:
          if i == hdr.messagefrom: 
            continue
          destination = i 
          hdr1 = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.WAVE, self.componentinstancenumber, destination)
          wave_msg = GenericMessage(hdr1, payload)
          self.send_down(Event(self, EventTypes.MFRT, wave_msg))
          self.waitingAccepts.append(i)
      else: 
        destination = hdr.messagefrom
        hdr1 = ApplicationLayerMessageHeader(ApplicationLayerMessageTypes.ACCEPT_WAVE, self.componentinstancenumber, destination)
        wave_msg = GenericMessage(hdr1, payload)
        self.send_down(Event(self, EventTypes.MFRT, wave_msg))
          

  def __init__(self, componentname, componentinstancenumber):
    super().__init__(componentname, componentinstancenumber)

    self.eventhandlers["propose"] = self.on_propose
    self.eventhandlers["agree"] = self.on_agree
    self.eventhandlers["timerexpired"] = self.on_timer_expired
    self.parent = componentinstancenumber
    self.initiated = False
    self.isWaiting = False 

    self.waitingAccepts = []



import logging

logger = logging.getLogger(__name__)


class LogManager:
    """Instantiate and manage the components of the "Log" subsystem.
    That is: the log, the compactor and the state machine."""

    def __init__(self):
        self.log = []
        self.commitIndex = len(self.log)

    def __getitem__(self, index):
        """Get item or slice from the log, based on absolute log indexes.
        Item(s) already compacted cannot be requested."""
        if type(index) is slice:
            start = index.start
            stop = index.stop
            return self.log[start:stop:index.step]
        elif type(index) is int:
            return self.log[index]

    @property
    def index(self):
        """Log tip index."""
        return len(self.log)

    def term(self, index=None):
        """Return a term given a log index. If no index is passed, return
        log tip term."""
        if index is None:
            return self.term(self.index)
        elif index == -1:
            return 0
        elif not len(self.log):
            return 0
        else:
            return self.log[index]['term']

    def append_entries(self, entries, prevLogIndex):
        start = prevLogIndex
        if len(self.log) >= start:
            del self.log[:start]
            self.log.extend(entries)
        else:
            self.log.extend(entries)
        if entries:
            logger.debug('Appending. New log: %s', self.log)

    def commit(self, leaderCommit):
        if leaderCommit <= self.commitIndex:
            return

        self.commitIndex = min(leaderCommit, self.index)  # no overshoots
        logger.debug('Advancing commit to %s', self.commitIndex)
        # above is the actual commit operation, just incrementing the counter!



from ahc.Ahc import ComponentModel, Event, ConnectorTypes
from Routing.FredericksonAlgorithmSimple.RoutingExample.ApplicationComponent import ApplicationComponent
from Routing.FredericksonAlgorithmSimple.RoutingExample.CoordinatorComponent import Coordinator
from Routing.FredericksonAlgorithmSimple.RoutingExample.FredericksonAlgorithmSimpleComponent import FredericksonAlgorithmSimpleComponent
from Routing.FredericksonAlgorithmSimple.RoutingExample.MiddlewareStubComponent import MiddlewareMessageStub

# Encapsulator for the Application Node
class MachineLearningNode(ComponentModel):
    def __init__(self, componentname, componentid):
        super(MachineLearningNode, self).__init__(componentname, componentid)
        self.Application = ApplicationComponent(ApplicationComponent.__name__, componentid)
        self.Coordinator = Coordinator(Coordinator.__name__, componentid)
        self.FrederickAlgorithmSimpleComponent = FredericksonAlgorithmSimpleComponent(FredericksonAlgorithmSimpleComponent.__name__, componentid)
        self.MiddlewareStub = MiddlewareMessageStub(MiddlewareMessageStub.__name__, componentid)

        # Application layer only talks with Coordinator,
        # Coordinator coordinates all demands of the application by triggering other components if necessary...
        self.Application.connect_me_to_component(ConnectorTypes.DOWN, self.Coordinator)
        self.Coordinator.connect_me_to_component(ConnectorTypes.UP, self.Application)
        self.Coordinator.connect_me_to_component(ConnectorTypes.PEER, self.FrederickAlgorithmSimpleComponent)
        self.Coordinator.connect_me_to_component(ConnectorTypes.DOWN, self.MiddlewareStub)

        self.FrederickAlgorithmSimpleComponent.connect_me_to_component(ConnectorTypes.PEER, self.Coordinator)

        self.FrederickAlgorithmSimpleComponent.connect_me_to_component(ConnectorTypes.DOWN, self.MiddlewareStub)
        self.MiddlewareStub.connect_me_to_component(ConnectorTypes.UP, self.FrederickAlgorithmSimpleComponent)
        self.MiddlewareStub.connect_me_to_component(ConnectorTypes.UP, self.Coordinator)

        self.MiddlewareStub.connect_me_to_component(ConnectorTypes.DOWN, self)
        self.connect_me_to_component(ConnectorTypes.UP, self.MiddlewareStub)

    def on_init(self, eventobj: Event):
        pass

    def on_message_from_bottom(self, eventobj: Event):
        self.send_up(eventobj) # send incoming messages to upper components

    def on_message_from_top(self, eventobj: Event):
        self.send_down(eventobj) # send incoming messages from upper components to a channel




import os
import pickle
from timeit import default_timer as timer

def singleton(cls):
  instance = [None]
  def wrapper(*args, **kwargs):
    if instance[0] is None:
      instance[0] = cls(*args, **kwargs)
    return instance[0]
  return wrapper

@singleton
class ExperimentLogger:
  finished = False
  start_time = None
  finish_time = None
  total_hops = 0

  def __init__(self):
    print("Logger Created!")

  # setters
  def set_finished(self):
    self.finished = True
    print("Total route finding time: ", ExperimentLogger().get_time_elapsed())
    print("Total hops for route finding: ", ExperimentLogger().get_total_hops())
    self.storeResult()

  def set_start_time(self, s_time):
    self.start_time = s_time

  def set_finish_time(self, f_time):
    self.finish_time = f_time

  def set_total_hops(self, hops):
    self.total_hops = hops

  # getters
  def get_finished(self):
    return self.finished
  
  def get_time_elapsed(self):
    return self.finish_time - self.start_time

  def get_total_hops(self):
    return self.total_hops

  def storeResult(self):
    elapsed_time = self.finish_time - self.start_time 
    with open(os.getcwd()+".30_"+str(timer())+".exp", "wb") as handle:
      pickle.dump((10, elapsed_time, self.total_hops), handle, protocol=pickle.HIGHEST_PROTOCOL)

